# SPDX-FileCopyrightText: 2019-2024 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
Main program file for the `lctime` standard-cell characterization tool.
"""

import os
import argparse
import joblib
import tempfile
from collections import Counter
from typing import List, Set

import liberty.parser as liberty_parser
from liberty.types import *
from ..cell_types import Combinational, SingleEdgeDFF, Latch, CellType

from PySpice.Unit import *

from ..logic.util import is_unate_in_xi, find_unateness_conditions
from ..liberty import util as liberty_util
from ..logic import functional_abstraction
from ..logic import seq_recognition
from ..logic.types import CombinationalOutput

from .timing_combinatorial import characterize_comb_cell
from .timing_sequential import *
from .input_capacitance import characterize_input_capacitances
from .ngspice_subprocess import Simulator, NgSpiceInteractive, NgSpiceFileBased
from . import flipflop
from . import util

from copy import deepcopy

from ..lccommon import net_util
from ..lccommon.net_util import load_transistor_netlist, is_ground_net, is_supply_net
from ..lccommon.data_types import Transistor, ChannelType
import networkx as nx
from sympy.logic import satisfiable
import sympy.logic.boolalg

from PySpice.Spice.Parser import SpiceParser
import logging

from ..licence import licence_notice_string_single_line, original_source_repository


def _boolean_to_lambda(boolean: boolalg.Boolean, input_pins: List):
    """
    Convert a sympy.boolalg.Boolean expression into a Python lambda function.
    :param boolean: Boolean function
    :param input_pins: Ordered list of input pins. Used to keep ordering of arguments consistent between lambda function and original cell function.
    :return:
    """
    simple = sympy.simplify(boolean)
    atoms = boolean.atoms()
    input_pins = [sympy.Symbol(i) for i in input_pins]
    inputs = [i for i in input_pins if i in atoms]
    f = sympy.lambdify(inputs, simple)
    return f


def recognize_cell_from_liberty(cell_group: Group) -> Union[Latch, SingleEdgeDFF, Combinational]:
    """
    Analyze the liberty group and derive the type of the cell.
    :param cell_group:
    :return:
    """

    # Get information on pins from the liberty file.
    input_pins, outputs_user = liberty_util.get_pin_information(cell_group)

    ff_group = cell_group.get_groups("ff")
    latch_group = cell_group.get_groups("latch")

    if ff_group and latch_group:
        logger.error("Cell contains a 'ff' and 'latch' description.")
        assert False, "Cannot characterize cells with both 'ff' and 'latch'."
    elif ff_group:
        if len(ff_group) != 1:
            assert False, "Cannot characterize cells with more than one 'ff' group."
        logger.info("'ff' group found. Cell is expected to be a flip-flop.")
        ff_group = ff_group[0]
        assert isinstance(ff_group, Group)

        # Get state names.
        iq, iqn = ff_group.args
        clocked_on = ff_group.get_boolean_function('clocked_on')
        next_state = ff_group.get_boolean_function('next_state')

        clear = ff_group.get_boolean_function('clear')
        preset = ff_group.get_boolean_function('preset')

        # TODO: Store and use this.
        clear_preset_var1 = ff_group['clear_preset_var1']
        clear_preset_var2 = ff_group['clear_preset_var2']

        # TODO: Store and use this.
        clocked_on_also = ff_group.get_boolean_function('clocked_on_also')

        cell_type = SingleEdgeDFF()
        cell_type.internal_state = sympy.Symbol(iq)
        cell_type.internal_state_n = sympy.Symbol(iqn)
        if clocked_on is not None:
            cell_type.clocked_on = clocked_on
        if next_state is not None:
            cell_type.next_state = next_state
        if preset is not None:
            cell_type.async_preset = preset
        if clear is not None:
            cell_type.async_clear = clear

    elif latch_group:
        if len(latch_group) != 1:
            assert False, "Cannot characterize cells with more than one 'latch' group."
        logger.info("'latch' group found. Cell is expected to be a latch.")
        latch_group = latch_group[0]
        assert isinstance(latch_group, Group)

        # Get state names.
        iq, iqn = latch_group.args

        cell_type = Latch()
        cell_type.internal_state = sympy.Symbol(iq)

        clocked_on = latch_group.get_boolean_function('clocked_on')
        next_state = latch_group.get_boolean_function('next_state')

        clear = latch_group.get_boolean_function('clear')
        preset = latch_group.get_boolean_function('preset')

        # TODO: Store and use this.
        clear_preset_var1 = latch_group['clear_preset_var1']
        clear_preset_var2 = latch_group['clear_preset_var2']

        # TODO: Store and use this.
        clocked_on_also = latch_group.get_boolean_function('clocked_on_also')

        cell_type.clocked_on = clocked_on
        cell_type.next_state = next_state
        cell_type.async_preset = preset
        cell_type.async_clear = clear
    else:
        # No sequential element.
        cell_type = Combinational()

    # TODO:
    # power_down_function = cell_group.get_boolean_function('power_down_function')
    # cell_type.power_down_function = power_down_function

    # Copy description of output pins.
    for name, output in outputs_user.items():
        cell_type.outputs[sympy.Symbol(name)] = output

    input_pins = {sympy.Symbol(p) for p in input_pins}

    inputs_found_in_outputs = {i for output in cell_type.outputs.values()
                               for i in output.get_inputs()}
    diff = inputs_found_in_outputs - input_pins
    if diff:
        logger.warning(f"Some input pins are used but not declared in the liberty file: {sorted(diff)}")

    cell_type.inputs = sorted(input_pins, key=str)

    return cell_type


def abort(message: str, exit_code=1):
    """
    Exit the program due to an error.
    :param message: Error message.
    :param exit_code:
    """
    logger.error(message)
    exit(exit_code)


def main():
    """
    Command-line tool for cell characterization.
    Currently only combinatorial cells are supported excluding tri-state cells.
    :return:
    """

    print(licence_notice_string_single_line())

    lctime = LcTime()
    lctime.run()


class LcTime:

    def __init__(self):
        self.args = None
        self.workingdir: str = None
        self.library: Group = None
        self.new_library: Group = None # Library under construction.
        self.conf: CharacterizationConfig = None
        self.setup_statements = []

        self.simulators = [
            ('ngspice-basic', NgSpiceFileBased),
            ('ngspice-interactive', NgSpiceInteractive),
        ]
        """
        Mapping of simulator names to simulator classes.
        """

    def run(self):
        logger = logging.getLogger(__name__)
        logger.info("lctime main function")

        self.f00_parse_args()
        self.f01_init_logging()
        self.f02_init_workdir()
        self.f03_init_netlists()
        self.f04_check_cell_names()
        self.f05_load_liberty_template()
        self.f06_check_delay_model()
        self.f07_init_new_library()
        self.f08_load_operating_conditions()
        self.f09_init_include_statements()
        self.f10_init_characterization_config()
        self.f11_init_table_indices()
        self.f12_characterize_library()
        self.write_liberty()

    def f00_parse_args(self):
        parser = argparse.ArgumentParser(
            description='Characterize the timing of a combinatorial cell based on a SPICE netlist. '
                        'The resulting liberty file will contain the data of the input liberty file '
                        'plus the updated characteristics of the selected cell.',
            epilog='Example: lctime --liberty specification.lib --cell INVX1 AND2X1 --spice netlists.sp -I '
                   'transistor_model.m --output mylib.lib')

        parser.add_argument('-l', '--liberty', required=True, metavar='LIBERTY', type=str,
                            help='Liberty file. This must contain all necessary specifications '
                                 'needed to characterize the cell.')

        parser.add_argument('--cell', required=True, metavar='CELL_NAME', type=str,
                            action='append',
                            nargs='+',
                            help='Names of cells to be characterized.')

        parser.add_argument('--spice', required=True, metavar='SPICE', type=str,
                            action='append',
                            nargs='+',
                            help='SPICE netlist containing a subcircuit with the same name as the cell.')

        parser.add_argument('-I', '--include', required=False, action='append', metavar='SPICE_INCLUDE', type=str,
                            help='SPICE files to include such as transistor models.')

        parser.add_argument('-L', '--library', required=False, action='append', metavar='SPICE_LIB', type=str,
                            help='SPICE .LIB statements defining each a path to the library and a library name.'
                                 'Example: --library "/path/to/lib libraryName".')

        parser.add_argument('--calc-mode', metavar='CALC_MODE', type=str, choices=['worst', 'typical', 'best'],
                            default='typical',
                            help='Calculation mode for computing the default timing arc'
                                 ' based on the conditional timing arcs. "worst", "typical" (average) or "best".')

        parser.add_argument('-o', '--output', required=True, metavar='LIBERTY_OUT', type=str, help='Output liberty file.')

        parser.add_argument('--workingdir', required=False, metavar='WORKDIR', type=str,
                            help="Directory for ngspice simulation scripts and raw results.")

        parser.add_argument('--output-loads', required=True, metavar='CAPACITANCES', type=str,
                            help="List of output load capacitances for the cells. In pico Farads."
                                 " List must be quoted, elements must be separated by a comma."
                                 " Example: '0.05, 0.1, 0.2'")

        parser.add_argument('--slew-times', required=True, metavar='SLEWTIMES', type=str,
                            help="List of slew times of the input signals in nano seconds."
                                 " List must be quoted, elements must be separated by a comma."
                                 " Example: '0.05, 0.1, 0.2'")

        parser.add_argument('--related-pin-transition', required=False, metavar='SLEWTIMES', type=str,
                            help="List of slew times of the clock signal in nano seconds. "
                                 "This is used for sequential cells only. "
                                 "List must be quoted, elements must be separated by a comma. "
                                 "Example: '0.05, 0.1, 0.2'")

        parser.add_argument('--analyze-cell-function', action='store_true',
                            help='Derive the logical function of the cell from the SPICE netlist (experimental).')

        parser.add_argument('--diff', required=False,
                            nargs="+",
                            metavar='DIFFERENTIAL_PATTERN',
                            type=str,
                            help='Specify differential inputs as "NonInverting,Inverting" tuples.'
                                 'The placeholder "%%" can be used like "%%_P,%%_N" or "%%,%%_Diff", ...')

        parser.add_argument('--time-step', default=10e-12,
                            metavar='TIME_STEP',
                            type=float,
                            help='Specify the simulation time-step in seconds. Default is 10e-12.')

        parser.add_argument('--debug', action='store_true',
                            help='Enable debug mode (more verbose logging).')

        parser.add_argument('--debug-plots', action='store_true',
                            help='Create debug plots of simulation waveforms.')

        parser.add_argument('-j', '--jobs', type=int, default=1, metavar="NUM_JOBS", help="number of threads for cell-level parallelism")

        parser.add_argument("--simulator",
                            choices=[n for n, _ in self.simulators],
                            default=self.simulators[0][0],
                            help="circuit simulator backend to be used"
                        )

        # Parse arguments
        self.args = parser.parse_args()
        

    def f01_init_logging(self):
        DEBUG = self.args.debug
        log_level = logging.DEBUG if DEBUG else logging.INFO

        if DEBUG:
            # Also output name of function in DEBUG mode.
            log_format = '%(module)16s %(funcName)16s %(levelname)8s: %(message)s'
        else:
            log_format = '%(levelname)8s: %(message)s'

        logging.basicConfig(format=log_format, level=log_level)
    
    def f02_init_workdir(self):
        self.workingdir = self.args.workingdir
        if self.workingdir is None:
            self.workingdir = tempfile.mkdtemp(prefix="lctime-")

    def f03_init_netlists(self):
        
        # Get list of cell names to be characterized.
        self.cell_names = [n for names in self.args.cell for n in names]  # Flatten the nested list.

        # Get list of user-provided netlist files.
        netlist_files = [n for names in self.args.spice for n in names]  # Flatten the nested list.

        # Generate a lookup-table which tells for each cell name which netlist file to use.
        self.netlist_file_table: Dict[str, str] = dict()
        for netlist_file in netlist_files:
            logger.info("Load SPICE netlist: {}".format(netlist_file))
            parser = SpiceParser(path=netlist_file)
            for sub in parser.subcircuits:
                if sub.name in self.netlist_file_table:
                    # Abort if a sub circuit is defined in multiple netlists.
                    abort(
                        f"Sub-circuit '{sub.name}' is defined in multiple netlists: {self.netlist_file_table[sub.name]}, {netlist_file}")
                self.netlist_file_table[sub.name] = netlist_file

    def f04_check_cell_names(self):
        # Test if all cell names can be found in the netlist files.
        cell_names_not_found = set(self.cell_names) - self.netlist_file_table.keys()
        if cell_names_not_found:
            abort(f"Cell names not found in netlists: {', '.join(cell_names_not_found)}")

    def f05_load_liberty_template(self):
        # Load liberty file.
        lib_file = self.args.liberty
        logger.info("Reading liberty: {}".format(lib_file))
        with open(lib_file) as f:
            data = f.read()
        self.library = liberty_parser.parse_liberty(data)

    def f06_check_delay_model(self):
        # Check if the delay model is supported.
        delay_model = self.library['delay_model']
        supported_delay_models = ['table_lookup']
        if delay_model not in supported_delay_models:
            msg = "Delay model not supported: '{}'. Must be one of {}.".format(delay_model,
                                                                               ", ".join(supported_delay_models))
            logger.debug(msg)
            abort(msg)

    def f07_init_new_library(self):
        # Make independent copies of the library.
        self.new_library = deepcopy(self.library)
        # Strip all cell groups.
        self.new_library.groups = [g for g in self.new_library.groups if g.group_name != 'cell']
        # Strip away all LUT template table groups.
        table_types = ['lu_table_template', 'power_lut_template']
        self.new_library.groups = [g for g in self.new_library.groups if g.group_name not in table_types]

    def f10_init_characterization_config(self):
        
        # Setup configuration struct.
        conf = CharacterizationConfig()
        
        # Load operation voltage and temperature.
        # TODO: load voltage/temperature from operating_conditions group
        conf.supply_voltage = self.library['nom_voltage']
        logger.info('Supply voltage = {:f} V'.format(conf.supply_voltage))
        conf.temperature = self.library['nom_temperature']
        logger.info('Temperature = {:f} V'.format(conf.temperature))

        # Read trip points from liberty file.
        conf.trip_points = util.read_trip_points_from_liberty(self.library)
        logger.debug(conf.trip_points)
        
        # Get timing corner from liberty file.
        # TODO: let user overwrite it.
        calc_modes = {
            'typical': CalcMode.TYPICAL,
            'worst': CalcMode.WORST,
            'best': CalcMode.BEST,
        }
        assert self.args.calc_mode in calc_modes, "Unknown calculation mode: {}".format(self.args.calc_mode)
        conf.timing_corner = calc_modes[self.args.calc_mode]
        logger.info("timing corner: {}".format(conf.timing_corner.name))
        
        conf.setup_statements = self.setup_statements
        
        # Simulation time step
        time_resolution_seconds = float(self.args.time_step)
        logger.info(f"Time resolution = {time_resolution_seconds}s")
        if time_resolution_seconds <= 0:
            abort('Time step must be larger than zero.')

        if time_resolution_seconds > 1e-9:
            logger.warning(f"Timestep is larger than 1ns: {time_resolution_seconds}s")
        conf.time_step = time_resolution_seconds
        
        conf.workingdir = self.workingdir
        conf.debug = self.args.debug
        conf.debug_plots = self.args.debug_plots
        conf.get_units_from_liberty(self.library)

        logger.info(f"Capacitance unit: {conf.capacitance_unit} F")
        logger.info(f"Time unit: {conf.time_unit} s")
        self.conf = conf
        
    def f08_load_operating_conditions(self):
        # TODO: not used at the moment
        # Get timing corner from liberty file.
        # Find definitions of operating conditions and sort them by name.
        operating_conditions_list = self.library.get_groups('operating_conditions')
        # Put into a dict by name.
        operating_conditions: Dict[str, Group] = {str(g.args[0]): g for g in operating_conditions_list}

        logger.info("Operating conditions: {}".format(set(operating_conditions.keys())))

        """
        TODO: Use the information from the operating conditions.
        Example:
        operating_conditions (MPSS) {
            calc_mode : worst ;
            process : 1.5 ;
            process_label : "ss" ;
            temperature : 70 ;
            voltage : 4.75 ;
            tree_type : worse_case_tree ;
        }
        """

        # TODO: Make use of this.
        default_operating_conditions = self.library['default_operating_conditions']
        logger.info("Default operating conditions: {}".format(default_operating_conditions))
        
    def f09_init_include_statements(self):
        # Create .include statements for ngspice
        
        spice_includes = self.args.include if self.args.include else []
        if len(spice_includes) == 0:
            logger.warning("No transistor model supplied. Use --include or -I.")

        # Sanitize include paths.
        input_argument_error = False
        for path in spice_includes:
            if not os.path.isfile(path):
                logger.error(f"Include file does not exist: {path}")
                input_argument_error = True

        spice_libraries_raw: List[str] = self.args.library if self.args.library else []
        # Split library statements into path and library name.
        spice_libraries: List[Tuple[str, str]] = [tuple(s.strip() for s in l.split(" ", maxsplit=2))
                                                  for l in spice_libraries_raw
                                                  ]
        # Sanitize the library arguments.
        for lib, raw in zip(spice_libraries, spice_libraries_raw):
            if len(lib) != 2 or not lib[0] or not lib[1]:
                abort(f'Library statements must be of the format "/path/to/library libraryName". Found: "{raw}".')

            path, name = lib
            if not os.path.isfile(path):
                logger.error(f"Library file does not exist: {path}")
                input_argument_error = True

        # Exit if some input arguments were obviously invalid.
        if input_argument_error:
            abort("Exit because of invalid arguments.")

        # .LIB statements
        library_statements = [f".LIB {path} {name}" for path, name in spice_libraries]

        # .INCLUDE statements
        include_statements = [f".include {i}" for i in spice_includes]

        self.setup_statements = library_statements + include_statements
        
    def f11_init_table_indices(self):
        """
        Load values for input transition times and output loads.
        """
        
        # Setup array of output capacitances and input slews.
        self.output_capacitances = np.array([float(s.strip()) for s in self.args.output_loads.split(",")]) * 1e-12  # pF
        self.input_transition_times = np.array([float(s.strip()) for s in self.args.slew_times.split(",")]) * 1e-9  # ns

        # Transition times of the clock pin.
        if self.args.related_pin_transition:
            self.related_pin_transition = np.array(
                [float(s.strip()) for s in self.args.related_pin_transition.split(",")]) * 1e-9  # ns
        else:
            self.related_pin_transition = None

        logger.info(f"Output capacitances [pF]: {self.output_capacitances / 1e-12}")
        logger.info(f"Input slew times [ns]: {self.input_transition_times / 1e-9}")
        if self.related_pin_transition is not None:
            logger.info(f"Related pin transition times [ns]: {self.related_pin_transition / 1e-9}")

    def f12_characterize_library(self):
        if self.args.jobs != 1:
            # Characterize cells in parallel.
            new_cell_groups = joblib.Parallel(n_jobs=self.args.jobs, prefer='threads') \
                (joblib.delayed(self.characterize_cell)(cell_name) for cell_name in self.cell_names)
        else:
            # Characterize cells sequentially.
            new_cell_groups = [self.characterize_cell(cell_name) for cell_name in self.cell_names]

        for new_cell_group in new_cell_groups:
            self.new_library.groups.append(new_cell_group)      


    def write_liberty(self):
        """
        Dump new liberty library to disk.
        """
        assert self.new_library is not None
        
        logger.info("Write liberty: {}".format(self.args.output))
        with open(self.args.output, 'w') as f:

            # Write URL to lctime.

            s = """/*
Generated by lctime ({}).
*/
    
    """
            f.write(s.format(original_source_repository()))

            f.write(str(self.new_library))
        
    def characterize_cell(self, cell_name: str) -> Group:
        """
        Characterize a cell and create an updated cell group.
        :param cell_name:
        :return: Return an updated cell group.
        """

        cell_setup = CellSetup(self, cell_name)

        return cell_setup.run()

class CellSetup:
    """
    Measurement setup for a single cell.
    """

    def __init__(self, lctime: LcTime, cell_name: str):
        self.lctime = lctime
        """
        Global values.
        """
        self.cell_name: str = cell_name
        self.cell_workingdir: str = None
        self.output_capacitances = None
        self.input_transition_times = None
        self.related_pin_transition = None
        self.cell_group = None
        self.cell_type_liberty = None
        "Cell type as derived from liberty file."
        self.cell_type = None
        self.transistors_abstract: List[Transistor] = None
        "List of transistors which make the circuit of the cell"
        self.well_pins: Dict[str, str] = dict()
        "power/ground/well nets"

    def init_workdir(self):
        """
        Create working directory if it does not exist yet.
        """
        self.cell_workingdir = os.path.join(self.lctime.conf.workingdir, self.cell_name)
        if not os.path.exists(self.cell_workingdir):
            os.mkdir(self.cell_workingdir)

    def init_indices(self):
        self.output_capacitances = self.lctime.output_capacitances
        self.input_transition_times = self.lctime.input_transition_times
        self.related_pin_transition = self.lctime.related_pin_transition

    def recognize_cell_type(self):
        try:
            self.cell_group = select_cell(self.lctime.library, self.cell_name)

            # Determine type of cell (latch, flip-flop, combinational).
            self.cell_type_liberty = recognize_cell_from_liberty(self.cell_group)
            self.cell_type = self.cell_type_liberty
        except KeyError as e:
            logger.warning(f"No cell group defined yet in liberty file: {self.cell_name}")

            if not self.lctime.args.analyze_cell_function:
                abort("Cell is not defined in liberty. Enable cell recognition with --analyze.")

            self.cell_type_liberty = Combinational()  # Default: Empty cell.

            # Cell group does not exist, so create it.
            logger.debug("Create empty cell group.")
            self.cell_group = Group(group_name='cell', args=[self.cell_name])
            self.lctime.library.groups.append(self.cell_group)

            # The liberty did not define anything about this cell.
            self.cell_type = None

        # Check that the name matches.
        assert self.cell_group.args == [self.cell_name], "Cell name does not match."  # This should not happen.
        
    def init_pins(self):
        # Get information on pins from the liberty file.
        self.input_pins = [str(s) for s in self.cell_type_liberty.inputs]
        self.output_pins = [str(s) for s in self.cell_type_liberty.outputs.keys()]
        
        logger.info(f"Input pins as defined in liberty: {self.input_pins}")
        logger.info(f"Output pins as defined in liberty: {sorted(self.cell_type_liberty.outputs.keys())}")


    def init_case_lookup_table(self):
        """
        Create lookup table to fix lower/upper case mismatches between Liberty and SPICE.
        """
        
        # Get all IO pins defined in liberty.
        liberty_pins = set(self.input_pins) | set(self.output_pins)
        # Convert to strings.
        liberty_pins = {str(p) for p in liberty_pins}
        
        # Create a lookup table to reconstruct lower/upper case letters.
        # This is a workaround. The SPICE parser converts everything to uppercase.
        self.case_lookup_table = {p.lower(): p for p in liberty_pins}
        if len(self.case_lookup_table) != len(liberty_pins):
            # It's not a one-to-one mapping!
            logger.warning(f"Mixed lower case and upper case could cause trouble.")

    def fix_case(self, pin: str) -> str:
        """
        Restore lower/upper case of signals that went lost during SPICE parsing.
        """
        return self.case_lookup_table.get(pin.lower(), pin)

    def init_differential_inputs(self):

        # Extract differential pairs from liberty.
        # Liberty allows to reference complementary pins with the 'complementary_pin' attribute.
        logger.debug("Load complementary pins from liberty.")
        differential_inputs_liberty = dict()
        for pin in self.cell_group.get_groups("pin"):
            assert isinstance(pin, liberty_parser.Group)
            pin_name = pin.args[0]
            complementary_pin = pin.get("complementary_pin")
            if complementary_pin is not None:
                differential_inputs_liberty[pin_name] = complementary_pin

        # Match differential inputs.
        if self.lctime.args.diff is not None:
            logger.debug("Match complementary pins from user-defined pattern.")
            differential_inputs_from_pattern = find_differential_inputs_by_pattern(self.lctime.args.diff, self.input_pins)
        else:
            differential_inputs_from_pattern = dict()

        # Merge the differential pairs provided from the user with the pairs detected from the liberty file.
        differential_inputs_liberty.update(differential_inputs_from_pattern)
        self.differential_inputs = differential_inputs_liberty

        # Sanity checks on complementary pins.
        # Complementary pin should not be defined as pin group in liberty file.
        for pin in self.cell_group.get_groups("pin"):
            assert isinstance(pin, liberty_parser.Group)
            pin_name = pin.args[0]
            if pin_name in self.differential_inputs.values():
                logger.warning(
                    f"Complementary pin is modelled in the liberty file but will not be characterized: {pin_name}")

        for noninv, inv in self.differential_inputs.items():
            logger.info(f"Differential input (+,-): {noninv}, {inv}")

        # Find all input pins that are not inverted inputs of a differential pair.
        inverted_pins = self.differential_inputs.values()

        self.input_pins_non_inverted = [p for p in self.input_pins if p not in inverted_pins]
        "All input pins that are not inverted inputs of a differential pair"


    def create_cell_config(self) -> CellConfig:
        # Setup cell specific configuration.
        cell_conf = CellConfig()
        cell_conf.cell_name = self.cell_name
        cell_conf.global_conf = self.lctime.conf
        cell_conf.complementary_pins = self.differential_inputs
        cell_conf.ground_net = self.gnd_pin()
        cell_conf.supply_net = self.vdd_pin()
        cell_conf.pwell_pin = self.pwell_pin()
        cell_conf.nwell_pin = self.nwell_pin()
        cell_conf.workingdir = self.cell_workingdir
        cell_conf.spice_netlist_file = self.lctime.netlist_file_table[self.cell_name]
        
        # Get pin ordering of spice circuit.
        cell_conf.spice_ports = get_subcircuit_ports(self.netlist_file_path(), self.cell_name)
        logger.debug(f"SPICE subcircuit ports: {cell_conf.spice_ports}")

        return cell_conf
        
    def netlist_file_path(self) -> str:
        """
        Get path to SPICE netlist.
        """
        return self.lctime.netlist_file_table[self.cell_name]

    def get_pg_pin_name_from_liberty(self, pg_type: str) -> str:
        """
        Get the specified power pin from the liberty structure, if present.
        """
        for pg_pin in self.cell_group.get_groups("pg_pin"):
            if pg_pin.get("pg_type") == pg_type:
                if len(pg_pin.args) != 1:
                    raise Exception(f"pg_pin group must have 1 argument but has {len(pg_pin.args)}")
                return pg_pin.args[0]
        return None
        
    def vdd_pin(self):
        # Get the primary power pin from the liberty, if present.
        p = self.get_pg_pin_name_from_liberty("primary_power")
        if p:
            return p

        # TODO: don't decide based only on net name.
        power_pins = self.power_pins()
        assert len(power_pins) == 2, "Expected to have 2 power pins."
        vdd_pins = [p for p in power_pins if net_util.is_supply_net(p)]
        assert len(vdd_pins) == 1, "Expected to find one VDD pin but found: {}".format(vdd_pins)
        return vdd_pins[0]
    
    def gnd_pin(self):
        # Get the primary ground pin from the liberty, if present.
        p = self.get_pg_pin_name_from_liberty("primary_ground")
        if p:
            return p
        
        # TODO: don't decide based only on net name.
        power_pins = self.power_pins()
        assert len(power_pins) == 2, "Expected to have 2 power pins."
        gnd_pins = [p for p in power_pins if net_util.is_ground_net(p)]
        assert len(gnd_pins) == 1, "Expected to find one GND pin but found: {}".format(gnd_pins)
        return gnd_pins[0]

    def pwell_pin(self):
        # Get the pwell pin from the liberty, if present.
        p_lib = self.get_pg_pin_name_from_liberty("pwell")
        p_rec = self.well_pins.get("pwell")
        p = p_rec
        if p_lib:
            if p_rec and p_rec != p_lib:
                logger.warning(f"recognized pwell pin {p_rec} does not match the pwell pin defined in the liberty file {p_lib}")
            p = p_lib

        if p == self.gnd_pin():
            # pwell is connected via VDD pin
            return None

        if p not in self.cell_pins:
            # pwell pin might be defined in liberty file
            # but not actually used in the cell.
            return None

        return p

    def nwell_pin(self):
        # Get the nwell pin from the liberty, if present.
        p_lib = self.get_pg_pin_name_from_liberty("nwell")
        p_rec = self.well_pins.get("nwell")
        p = p_rec
        if p_lib:
            if p_rec and p_rec != p_lib:
                logger.warning(f"recognized nwell pin {p_rec} does not match the nwell pin defined in the liberty file {p_lib}")
            p = p_lib

        if p == self.vdd_pin():
            # nwell is connected via VDD pin
            return None

        if p not in self.cell_pins:
            # nwell pin might be defined in liberty file
            # but not actually used in the cell.
            return None
        
        return p
        
        
    def power_pins(self) -> List[str]:
        # TODO: don't decide based only on net name.
        return [p for p in self.cell_pins if net_util.is_power_net(p)]
        

    def get_liberty_pins(self) -> Set:
        """
        Get set of pin names defined in liberty template.
        """
        liberty_pins = set()
        for pin in self.cell_group.get_groups("pin"):
            assert isinstance(pin, liberty_parser.Group)
            pin_name = pin.args[0]
            liberty_pins.add(pin_name)
            complementary_pin = pin.get("complementary_pin")
            if complementary_pin is not None:
                liberty_pins.add(complementary_pin)
        return liberty_pins
        
    def check_pins(self):
        """
        Sanity check: All pins defined in liberty must appear in the SPICE netlist.
        """
        all_spice_pins = set(self.cell_pins)
        pins_not_in_spice = sorted(self.get_liberty_pins() - all_spice_pins)

        if pins_not_in_spice:
            abort(f"Pins defined in liberty but not in SPICE netlist: {', '.join(pins_not_in_spice)}")

    def io_pins(self):
        """
        Get data pins, i.e. all pins which are not power pins.
        """
        # TODO: replace this heuristic by something more thorough
        return net_util.get_io_pins(self.cell_pins)

    def load_netlist(self):
        """ Load netlist of cell
        Populate the self.transistors_abstract variable.
        """
        # TODO: Load all netlists at the beginning of lctime to detect errors early.
        # Get netlist and liberty group.
        logger.info('Load netlist: %s', self.netlist_file_path())
        try:
            transistors_abstract, cell_pins = load_transistor_netlist(self.netlist_file_path(), self.cell_name)
            logger.debug(f"Pins of cell '{self.cell_name}': {cell_pins}")
        except Exception as e:
            abort(str(e))

            
        if len(transistors_abstract) == 0:
            msg = "No transistors found in cell. (The netlist must be flattened, sub-circuits are not resolved)"
            if self.lctime.args.analyze_cell_function:
                # Transistors are needed for cell recognition only.
                abort(msg)
            else:
                logger.info(f"{msg}. Characterization might work as well but recognition won't work.")

        self.cell_pins = [self.fix_case(p) for p in cell_pins]
        for t in transistors_abstract:
            t.source_net = self.fix_case(t.source_net)
            t.drain_net = self.fix_case(t.drain_net)
            t.gate_net = self.fix_case(t.gate_net)
            t.body_net = self.fix_case(t.body_net)

        self.transistors_abstract = transistors_abstract

    def _recognize_well_pins(self) -> Dict[str, str]:
        """
        Detect nwell/pwell pins from the transistor network.
        """
        assert self.transistors_abstract is not None
        
        all_spice_pins = set(self.cell_pins)
        
        nmos_body = Counter(t.body_net for t in self.transistors_abstract if t.channel_type == ChannelType.NMOS)
        pmos_body = Counter(t.body_net for t in self.transistors_abstract if t.channel_type == ChannelType.PMOS)

        nets = {"nwell": None, "pwell": None}

        # Take only nets which are also a pin.
        nmos_body = [n for n, _ in nmos_body.most_common() if n in all_spice_pins]
        pmos_body = [n for n, _ in pmos_body.most_common() if n in all_spice_pins]
        
        if pmos_body:
            nwell_net = pmos_body[0]
            logger.info(f"Detected nwell pin: {nwell_net}")
            nets["nwell"] = nwell_net

        if nmos_body:
            pwell_net = nmos_body[0]
            logger.info(f"Detected pwell pin: {pwell_net}")
            nets["pwell"] = pwell_net

        return nets

    def _transistor_graph(self) -> nx.MultiGraph:
        """
        Get the network of transistors in the form of a networkx graph.

        Convert the transistor network into its multi-graph representation.
        This is used for a formal analysis of the network.
        """
        assert self.transistors_abstract is not None
        return _transistors2multigraph(self.transistors_abstract)

    def recognize_input_pins(self):
        """
        Dectect input pins / gates by analyzing the transistor network.
        """


        logger.debug("Detect input nets from the circuit.")
        detected_inputs = functional_abstraction.find_input_gates(self._transistor_graph())
        # Detect nets that must be inputs (connected to gates only but do not
        # appear in the list of pins in the SPICE circuit definition.
        all_spice_pins = set(self.cell_pins)
        inputs_missing_in_spice = detected_inputs - all_spice_pins
        if inputs_missing_in_spice:
            logger.warning(f"The circuit has gate nets that must be inputs "
                           f"but are not in the pin definition of the SPICE circuit: "
                           f"{', '.join(sorted(inputs_missing_in_spice))}")
        # Same check for pins declared in liberty template.
        inputs_missing_in_liberty = detected_inputs - self.get_liberty_pins()
        if inputs_missing_in_liberty:
            logger.warning(f"The circuit has gate nets that must be inputs "
                           f"but are not declared as a pin in the liberty template: "
                           f"{', '.join(sorted(inputs_missing_in_liberty))}")

        # Add detected input pins.
        diff = detected_inputs - set(self.input_pins)
        if diff:
            logger.info(f"Also include detected pins: {', '.join(sorted(diff))}")
            self.input_pins.extend(diff)

        power_pins = {self.vdd_pin(), self.gnd_pin(), self.nwell_pin(), self.pwell_pin()}

        # Find pins that are defined in the SPICE circuit but are not inputs nor power.
        maybe_outputs = all_spice_pins - set(self.input_pins) - power_pins - set(self.output_pins)
        if maybe_outputs:
            logger.info(f"Potential output pins: {', '.join(sorted(maybe_outputs))}")
            self.output_pins.extend(maybe_outputs)

    def recognize_cell_function(self):
            """
            Derive boolean functions for the outputs from the netlist.
            """
            logger.info("Derive boolean functions for the outputs based on the netlist.")

            self.cell_type = analyze_boolean_functions(
                self.cell_type,
                self.cell_group,
                self._transistor_graph(),
                self.io_pins(),
                self.differential_inputs,
                self.vdd_pin(),
                self.gnd_pin(),
            )

            merge_cell_types(self.cell_type_liberty, self.cell_type)

    def _create_simulation(
        self,
        simulator: Simulator,
    ) -> CellSimulation:
        """
        Setup the basic simulation harness.
        """

        cell_conf = self.create_cell_config()

        input_voltages = dict()
        input_currents = dict()

        for pin in self.input_pins:
            input_voltages[pin] = 0.0
                
        input_voltages[cell_conf.supply_net] = cell_conf.global_conf.supply_voltage
        if cell_conf.pwell_pin:
            input_voltages[cell_conf.pwell_pin] = 0.0
        if cell_conf.nwell_pin:
            input_voltages[cell_conf.nwell_pin] = cell_conf.global_conf.supply_voltage
    
        # Create a list of include files.
        setup_statements = cell_conf.global_conf.setup_statements + [f".include {cell_conf.spice_netlist_file}"]
    
        # Create simulation harness.
        simulation = CellSimulation(
            simulator,
            cell_name=cell_conf.cell_name,
            cell_ports=cell_conf.spice_ports,
            input_voltages=input_voltages, # First set all to 0V.
            input_currents=input_currents,
            output_load_capacitances={p: 0.0 for p in self.output_pins}, # TODO
            setup_statements=setup_statements,
            ground_net=cell_conf.ground_net,
            simulation_title=f"characterization (combinational) of {cell_conf.cell_name}"
        )

        return simulation

    def run(self) -> Group:
        """
        Characterize the cell and create a populated 'cell' group data structure.
        """
        
        logger.info("Cell: {}".format(self.cell_name))

        self.init_workdir()
        self.init_indices()
        self.recognize_cell_type()
        self.init_pins()
        self.init_case_lookup_table()
        self.load_netlist()
        self.check_pins()

        # Detect input nets from the transistor netlist (if enabled).
        if self.lctime.args.analyze_cell_function:
            self.well_pins = self._recognize_well_pins()
            self.recognize_input_pins()
            
        # Sanity check.
        if len(self.input_pins) == 0:
            # TODO: could be valid for tie cells.
            abort("Cell has no input pins.")

        # Sanity check.
        if len(self.output_pins) == 0:
            abort("Cell has no output pins.")

        self.init_differential_inputs()

        if self.lctime.args.analyze_cell_function:
            self.recognize_cell_function()
        else:
            # Skip functional abstraction and take the functions provided in the liberty file.
            # output_functions_symbolic = output_functions_user
            pass  # TODO


        logger.info(f"Run characterization for {self.cell_name}")

        cell_conf = self.create_cell_config()
        
        # Add groups for the cell to be characterized.
        new_cell_group = deepcopy(select_cell(self.lctime.library, self.cell_name))

        # Populate the cell group with missing pin groups.
        create_missing_pin_groups(
            new_cell_group,
            self.cell_type,
            cell_conf,
            self.output_pins,
            self.input_pins_non_inverted
        )

        # Specify power and ground pins.
        create_pg_pin_groups(
            new_cell_group,
            cell_conf,
        )

        # Get simulator backend.
        simulator_class = dict(self.lctime.simulators)[self.lctime.args.simulator]
        logger.debug(f"Backend class: {simulator_class.__name__}")
                
        with simulator_class(logger=logger) as simulator:
            logger.info(f"Backend: {simulator.name()}")

            # Measure input pin capacitances.
            measure_input_capacitances(
                simulator,
                self.lctime.conf,
                cell_conf,
                self.input_pins,
                self.input_pins_non_inverted,
                self.output_pins,
                new_cell_group
            )

            simulation = self._create_simulation(simulator)

            if isinstance(self.cell_type, Combinational):
                characterize_combinational_output(
                    simulation,
                    self.lctime.new_library,
                    new_cell_group,
                    self.lctime.conf,
                    self.cell_type,
                    cell_conf,
                    self.input_pins,
                    self.input_pins_non_inverted,
                    self.output_capacitances,
                    self.input_transition_times
                )
            elif isinstance(self.cell_type, SingleEdgeDFF):
                flipflop.characterize_flip_flop_output(
                    simulation,
                    self.lctime.new_library,
                    new_cell_group,
                    self.lctime.conf,
                    self.cell_type,
                    cell_conf,
                    self.related_pin_transition,
                    self.input_transition_times,
                    output_capacitances=self.output_capacitances
                )
            elif isinstance(self.cell_type, Latch):
                logger.info("Characterize latch.")
                logger.error("Characterization of latches is not yet supported.")

                """
                Characterization of latches is very similar to the one of flip-flops. Delays, hold and setup times
                are measured relative to the de-activating edge of the clock signal instead of the active clock edge.
                """

                abort("Characterization of latches is not yet supported.")
            else:
                assert False, f"Unsupported cell type: {type(self.cell_type)}"

        assert isinstance(new_cell_group, Group)
        return new_cell_group


def analyze_boolean_functions(
        cell_type: Optional[CellType],
        cell_group: Group,
        transistor_graph: nx.MultiGraph,
        io_pins: Set,
        differential_inputs: Dict[str, str],
        vdd_pin: str,
        gnd_pin: str,
) -> CellType:
    assert isinstance(cell_group, Group)
    assert cell_group.group_name == 'cell', "not a cell group"
    assert cell_type is None or isinstance(cell_type, CellType)

    abstracted_circuit = functional_abstraction.analyze_circuit_graph(
        graph=transistor_graph,
        pins_of_interest=io_pins,
        constant_input_pins={
            vdd_pin: True,
            gnd_pin: False},
        differential_inputs=differential_inputs,
        user_input_nets=None
    )

    if abstracted_circuit.latches:
        # There's some feedback loops in the circuit.

        # Try to recognize sequential cells.
        detected_cell_type = seq_recognition.extract_sequential_circuit(abstracted_circuit)

        if detected_cell_type:
            logger.info(f"Detected sequential circuit:\n{detected_cell_type}")

    else:
        logger.info("Detected purely combinational circuit.")
        detected_cell_type = Combinational()
        detected_cell_type.outputs = abstracted_circuit.outputs
        detected_cell_type.inputs = abstracted_circuit.get_primary_inputs()
        if cell_type is None:
            cell_type = Combinational()

    if cell_type is None or len(cell_group.groups) == 0:
        cell_type = detected_cell_type
    else:
        # Sanity check: Detected cell type (combinational, latch, ff) must match with the liberty file.
        if type(detected_cell_type) is not type(cell_type):
            msg = f"Mismatch: Detected cell type is {type(detected_cell_type)} " \
                  f"but liberty says {type(cell_type)}."
            logger.error(msg)
            assert False, msg

    output_functions_deduced = abstracted_circuit.outputs

    # Log deduced output functions.
    for output_name, value in output_functions_deduced.items():
        if value.high_impedance:
            logger.info(
                f"Deduced output function: {output_name} = {value.function}, tri_state = {value.high_impedance}")
        else:
            logger.info(f"Deduced output function: {output_name} = {value.function}")

    # # Convert keys into strings (they are `sympy.Symbol`s now)
    # output_functions_deduced = {str(output.name): comb.function for output, comb in
    #                             output_functions_deduced.items()}
    # output_functions_symbolic = output_functions_deduced.copy()

    return cell_type


def merge_cell_types(
        source_cell_type: CellType,
        target_cell_type: CellType
):
    # Merge deduced output functions with the ones read from the liberty file and perform consistency check.
    for output_symbol, output in source_cell_type.outputs.items():
        output_name = str(output_symbol)
        logger.info(f"User supplied output function: {output_name} = {output.function}")

        assert output_symbol in target_cell_type.outputs, f"No function has been deduced for output pin '{output_name}'."
        # Consistency check:
        # Verify that the deduced output formula is equal to the one defined in the liberty file.
        logger.info("Check equality of boolean function in liberty file and derived function.")
        equal = functional_abstraction.bool_equals(output.function,
                                                   target_cell_type.outputs[output_symbol].function)
        if not equal:
            msg = "User supplied function does not match the deduced function for pin '{}'".format(output_name)
            logger.error(msg)

        if equal:
            # Take the function defined by the liberty file.
            # This might be desired because it is in another form (CND, DNF,...).
            target_cell_type.outputs[output_symbol] = output


def create_missing_pin_groups(
        new_cell_group: Group,
        cell_type: CellType,
        cell_conf: CellConfig,
        output_pins: List[str],
        input_pins_non_inverted: List[str]
):
    # Strip away timing groups. They will be replaced by the new characterization.
    for pin_group in new_cell_group.get_groups('pin'):
        pin_group.groups = [g for g in pin_group.groups if g.group_name != 'timing']

    # Create missing pin groups.
    for pin in sorted(set(input_pins_non_inverted + output_pins)):
        pin_group = new_cell_group.get_groups('pin', pin)
        if not pin_group:
            pin_group = Group('pin', args=[pin])
            new_cell_group.groups.append(pin_group)

    # Set 'direction' attribute of input pins.
    for pin in input_pins_non_inverted:
        pin_group = new_cell_group.get_group('pin', pin)
        if 'direction' not in pin_group:
            pin_group['direction'] = 'input'

    # Set 'direction' attribute of output pins.
    for pin in output_pins:
        pin_group = new_cell_group.get_group('pin', pin)
        if 'direction' not in pin_group:
            pin_group['direction'] = 'output'
        pin_symbol = sympy.Symbol(pin)
        if cell_type.outputs[pin_symbol].is_tristate():
            # Mark as tri-state.
            pin_group.set_boolean_function('three_state', cell_type.outputs[pin_symbol].high_impedance)

    # Create 'complementary_pin' attribute for the inverted pin of differential pairs.
    for input_pin in input_pins_non_inverted:
        input_pin_group = new_cell_group.get_group('pin', input_pin)
        # Create link to inverted pin for differential inputs.
        input_pin_inverted = cell_conf.complementary_pins.get(input_pin)
        if input_pin_inverted:
            input_pin_group['complementary_pin'] = [EscapedString(input_pin_inverted)]


def create_pg_pin_groups(
        new_cell_group: Group,
        cell_conf: CellConfig,
):
    """
    Create the pg_pin groups which specify power/ground/pwell/nwell pins.
    """

    # TODO: What if this pg_pin group already exists?
    if not new_cell_group.get_groups("pg_pin", cell_conf.supply_net):
        primary_power = Group("pg_pin", args=[cell_conf.supply_net])
        primary_power["pg_type"] = EscapedString("primary_power")
        primary_power["voltage_name"] = EscapedString(cell_conf.supply_net)
        # TODO: primary_power["related_bias_pin"] = 
        new_cell_group.groups.append(primary_power)

    if not new_cell_group.get_groups("pg_pin", cell_conf.ground_net):
        primary_ground = Group("pg_pin", args=[cell_conf.ground_net])
        primary_ground["pg_type"] = EscapedString("primary_ground")
        primary_ground["voltage_name"] = EscapedString(cell_conf.ground_net)
        # TODO: primary_ground["related_bias_pin"] = 
        new_cell_group.groups.append(primary_ground)

    if not new_cell_group.get_groups("pg_pin", cell_conf.nwell_pin):
        nwell = Group("pg_pin", args=[cell_conf.nwell_pin])
        nwell["pg_type"] = EscapedString("nwell")
        nwell["voltage_name"] = EscapedString(cell_conf.nwell_pin)
        new_cell_group.groups.append(nwell)

    if not new_cell_group.get_groups("pg_pin", cell_conf.pwell_pin):
        pwell = Group("pg_pin", args=[cell_conf.pwell_pin])
        pwell["pg_type"] = EscapedString("pwell")
        pwell["voltage_name"] = EscapedString(cell_conf.pwell_pin)
        new_cell_group.groups.append(pwell)


def measure_input_capacitances(
        simulator: NgSpiceInteractive,
        conf: CharacterizationConfig,
        cell_conf: CellConfig,
        input_pins: List[str],
        input_pins_non_inverted: List[str],
        output_pins: List[str],
        cell_group: Group
):
    """
    Measure input capacitances and write them into the `cell_group`.
    :param conf: 
    :param cell_conf: 
    :param input_pins: 
    :param input_pins_non_inverted: 
    :param output_pins: 
    :param cell_group: 
    :return: 
    """
    logger.debug(f"Measuring input pin capacitances of cell {cell_conf.cell_name}.")
    for input_pin in input_pins_non_inverted:
        # Input capacitances are not measured for the inverting inputs of differential pairs.
        logger.info("Measuring input capacitance: {} {}".format(cell_conf.cell_name, input_pin))
        input_pin_group = cell_group.get_group('pin', input_pin)

        result = characterize_input_capacitances(
            simulator=simulator,
            input_pins=input_pins,
            active_pin=input_pin,
            output_pins=output_pins,
            cell_conf=cell_conf
        )

        input_pin_group['rise_capacitance'] = result['rise_capacitance'] / conf.capacitance_unit
        input_pin_group['fall_capacitance'] = result['fall_capacitance'] / conf.capacitance_unit
        input_pin_group['capacitance'] = result['capacitance'] / conf.capacitance_unit


def characterize_combinational_output(
        simulation: CellSimulation,
        new_library: Group,
        new_cell_group: Group,
        conf: CharacterizationConfig,
        cell_type: CellType,
        cell_conf: CellConfig,
        input_pins: List[str],
        input_pins_non_inverted: List[str],
        output_capacitances: np.ndarray,
        input_transition_times: np.ndarray
):
    assert isinstance(simulation, CellSimulation)
    assert isinstance(cell_type, Combinational) or isinstance(cell_type, SingleEdgeDFF)
    
    is_clock_to_output_arc = isinstance(cell_type, SingleEdgeDFF)
    
    # Measure timing for all input-output arcs.
    logger.debug("Measuring delay arcs.")
    for output_pin_symbol in cell_type.outputs.keys():
        output_pin_name = str(output_pin_symbol)
        output_pin_group = new_cell_group.get_group('pin', output_pin_name)

        # Store boolean function of output to liberty.
        output_pin_group.set_boolean_function('function', cell_type.outputs[output_pin_symbol].function)

        # Check if the output can be high-impedance.
        is_tristate = cell_type.outputs[output_pin_symbol].is_tristate()

        # Store three_state function to liberty.
        constant_input_pins = dict()
        if is_tristate:
            output_pin_group.set_boolean_function(
                'three_state', cell_type.outputs[output_pin_symbol].high_impedance
            )

            # Find normal operating conditions such that the output is not tri-state.
            models = list(satisfiable(~cell_type.outputs[output_pin_symbol].high_impedance, all_models=True))
            for model in models:
                logger.info(f"Output '{output_pin_name}' is low-impedance when {model}.")

            if len(models) > 1:
                abort("Characterization of tri-state outputs is not supported when the tri-state depends on"
                      "more than one input.")

            model = models[0]
            for name, value in model.items():
                constant_input_pins[str(name)] = value

        # Skip measuring the tri-state enable input.
        related_pins = sorted(set(input_pins_non_inverted) - constant_input_pins.keys())
        _input_pins = sorted(set(input_pins) - constant_input_pins.keys())

        # Characterize each from related_pin to output_pin.
        for related_pin in related_pins:

            related_pin_inverted = cell_conf.complementary_pins.get(related_pin)
            if related_pin_inverted:
                logger.info("Timing arc (differential input): ({}, {}) -> {}"
                            .format(related_pin, related_pin_inverted, output_pin_name))
            else:
                logger.info("Timing arc: {} -> {}".format(related_pin, output_pin_name))

            # Convert deduced output functions into Python lambda functions.
            output_functions = {
                str(name): _boolean_to_lambda(comb_output.function, input_pins)
                for name, comb_output in cell_type.outputs.items()
            }

            f = cell_type.outputs[output_pin_symbol].function

            # Find assignments of the other input pins such that
            # the output is positive unate / negative unate.
            pos_unate_values, neg_unate_values = find_unateness_conditions(f, related_pin)
            pos_unate_values = [v for v in pos_unate_values if v != False]
            neg_unate_values = [v for v in neg_unate_values if v != False]
            
            if pos_unate_values:
                logger.info(f"{related_pin}->{output_pin_name} is positive unate when {pos_unate_values}")
            if neg_unate_values:
                logger.info(f"{related_pin}->{output_pin_name} is negative unate when {neg_unate_values}")

            if not pos_unate_values and not neg_unate_values:
                # If this happens, then maybe something is wrong with the code here.
                log.warning(f"Timing arc {related_pin}->{output_pin_name} can never be activated.")

            timing_senses = {
                "positive_unate": pos_unate_values,
                "negative_unate": neg_unate_values,
            }

            # Consistency check.
            timing_sense = str(is_unate_in_xi(output_functions[output_pin_name], related_pin).name).lower()
            if pos_unate_values and neg_unate_values:
                assert timing_sense == 'non_unate'
            if pos_unate_values and not neg_unate_values:
                assert timing_sense == 'positive_unate'
            if not pos_unate_values and neg_unate_values:
                assert timing_sense == 'negative_unate'

            # Collect timing and power tables for all timing senses.
            results = dict()
            
            index_1 = None
            index_2 = None

            # Compute timing tables for all possible timing senses (positive/negative unate).
            for timing_sense, static_input_patterns in timing_senses.items():
                if not static_input_patterns:
                    # skip
                    continue

                # Convert pin names from sympy symbols to strings
                static_input_patterns = [
                    {str(k): v for k, v in pattern.items() if k != True}
                    for pattern in static_input_patterns
                ]

                logger.info("Timing sense: {}".format(timing_sense))

                result = characterize_comb_cell(
                    simulation=simulation,
                    input_pins=_input_pins,
                    output_pin=output_pin_name,
                    related_pin=related_pin,
                    output_functions=output_functions,

                    total_output_net_capacitance=output_capacitances,
                    input_net_transition=input_transition_times,

                    cell_conf=cell_conf,
                    constant_inputs=constant_input_pins,
                    static_input_patterns=static_input_patterns,
                )

                # Store result for computation of power tables.
                results[timing_sense] = result

                # Get the table indices.
                # TODO: get correct index/variable mapping from liberty file.
                index_1 = result['total_output_net_capacitance'] / conf.capacitance_unit
                index_2 = result['input_net_transition'] / conf.time_unit

                # Create template tables.
                delay_template_table = liberty_util.create_delay_template_table(new_library, len(index_1), len(index_2))
                delay_table_template_name = delay_template_table.args[0]

                # Create liberty timing tables.
                timing_tables = []
                for table_name in ['cell_rise', 'cell_fall', 'rise_transition', 'fall_transition']:
                    table = Group(
                        table_name,
                        args=[delay_table_template_name],
                    )

                    table.set_array('index_1', index_1)
                    table.set_array('index_2', index_2)
                    table.set_array('values', result[table_name] / conf.time_unit)

                    timing_tables.append(table)

                # Create the liberty timing group.
                timing_attributes = [
                    Attribute('related_pin', EscapedString(related_pin)),
                    Attribute('timing_sense', timing_sense)
                ]

                timing_group = Group(
                    'timing',
                    attributes=timing_attributes,
                    groups=timing_tables
                )

                # Attach timing group to output pin group.
                output_pin_group.groups.append(timing_group)

            # Create template tables.
            power_template_table = liberty_util.create_power_template_table(new_library, len(index_1), len(index_2))
            power_table_template_name = power_template_table.args[0]
            
            # Create liberty power tables.
            power_tables = []
            for table_name in ['rise_power', 'fall_power']:
                
                # Summarize power tables.
                power_tables_raw = []
                for result in results.values():
                    assert isinstance(result[table_name], np.ndarray) # Sanity check
                    power_tables_raw.append(result[table_name])

                power_table_raw = np.mean(np.array(power_tables_raw), axis=0) # TODO: it should be a weighted mean or something else
                    
                table = Group(
                    table_name,
                    args=[power_table_template_name],
                )

                assert index_1 is not None
                assert index_2 is not None
                assert power_table_raw.shape == (len(index_1), len(index_2)), "dimension mismatch"

                table.set_array('index_1', index_1)
                table.set_array('index_2', index_2)
                table.set_array('values', power_table_raw / conf.energy_unit)

                power_tables.append(table)

            power_attributes = [
                Attribute('related_pin', EscapedString(related_pin)),
            ]
            
            internal_power_group = Group(
                'internal_power',
                attributes=power_attributes,
                groups=power_tables
            )
            
            # Attach group to output pin group.
            output_pin_group.groups.append(internal_power_group)
    
def _transistors2multigraph(transistors) -> nx.MultiGraph:
    """ Create a graph representing the transistor network.
        Each edge corresponds to a transistor, each node to a net.
    """
    G = nx.MultiGraph()
    for t in transistors:
        G.add_edge(t.source_net, t.drain_net, (t.gate_net, t.channel_type))
    assert len(G) == 0 or nx.is_connected(G)
    return G




# Copyright (c) 2019-2020 Thomas Kramer.
# SPDX-FileCopyrightText: 2022 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
Simple sub-process based NgSpice binding.
"""

import os
import subprocess
import queue
import threading
import logging
import tempfile
import numpy as np
from typing import Dict, List, Tuple, Union, Optional
from .piece_wise_linear import PieceWiseLinear

logger = logging.getLogger(__name__)


def run_simulation(sim_file: str, ngspice_executable: str = 'ngspice'):
    """
    Invoke 'ngspice' to run the `sim_file`.
    :param sim_file: Path to ngspice simulation file.
    :return: Returns (stdout, stderr) outputs of ngspice.
    """
    logger.debug(f"Run simulation: {sim_file}")
    try:
        ret = subprocess.run([ngspice_executable, sim_file], capture_output=True)
        # proc = subprocess.Popen([ngspice_executable, sim_file])
        # logger.debug(f"Subprocess return value: {ret}")
        if ret.returncode != 0:
            ngspice_err_message = ret.stderr.decode("utf-8")
            logger.error(f"ngspice simulation failed: {ngspice_err_message}")
            raise Exception(f"ngspice simulation failed: {ngspice_err_message}")

        return ret.stdout.decode("utf-8"), ret.stderr.decode("utf-8")
    except FileNotFoundError as e:
        msg = f"SPICE simulator executable not found. Make sure it is in the current path: {ngspice_executable}"
        logger.error(msg)
        raise FileNotFoundError(msg)

class CellSimulation:

    def __init__(self,
        simulator, #: NgSpiceInteractive,
        cell_name: str,
        cell_ports: List[str],
        input_voltages: Dict[str, Union[float, PieceWiseLinear]],
        input_currents: Dict[str, Union[float, PieceWiseLinear]],
        output_load_capacitances: Dict[str, float] = None,
        setup_statements: List[str] = None,
        ground_net: str = "GND",
        simulation_title: str = "<UNTITLED SIMULATION>",
        initial_node_voltages: Dict[str, float] = None,
        set_initial_voltages_for_outputs: bool = True
    ):
        """
        :param simulator:
        :param cell_name: Name of the cell.
        :param cell_ports: List of all pins in the order they appear in the SPICE subcircuit.
        :param input_voltages: Values for driving input pins with a voltage source.
        :param input_currents: Values for driving input pins with a current source.
        :param output_load_capacitances: 
        :param setup_statements: ngspice statements such as 'include' or 'lib'
        :param ground_net: Name of the ground net.
        :param simulation_title: Descriptive title of the simulation.
        :param initial_node_voltages: Optionally provide initial node voltages here.
        :param set_initial_voltages_for_outputs: Enable initial node voltage hints for output pins

        """
        assert isinstance(simulator, Simulator)
        self.simulator = simulator
        self.cell_name = cell_name
        self.cell_ports = cell_ports
        self.input_voltages = input_voltages
        self.input_currents = input_currents
        self.initial_node_voltages = initial_node_voltages if initial_node_voltages else dict()
        self.simulation_title = simulation_title
        self.ground_net = ground_net
        self.temperature = 25
        self.output_load_capacitances = output_load_capacitances
        self.setup_statements = setup_statements
        self.set_initial_voltages_for_outputs = set_initial_voltages_for_outputs 
        missing_sources = set(cell_ports) - {ground_net} - input_voltages.keys() - input_currents.keys() - output_load_capacitances.keys()
        if missing_sources:
            logger.warning(f"missing sources for input pins: {missing_sources}")
            assert False

        self._init_simulation()

    def _initial_voltages(self) -> Dict[str, float]:
        initial_voltages = dict()
        for net, v in self.input_voltages.items():
            if isinstance(v, PieceWiseLinear):
                initial_voltages[net] = v(0)
            else:
                initial_voltages[net] = v
                
        # Set default value of outputs.
        if self.set_initial_voltages_for_outputs:
            for net in self.output_load_capacitances.keys():
                initial_voltages[net] = 0.0

        for net, v in self.initial_node_voltages.items():
            initial_voltages[net] = v
            
        return initial_voltages
        
    def _initial_currents(self) -> Dict[str, float]:
        initial_currents = dict()
        for net, v in self.input_currents.items():
            if isinstance(v, PieceWiseLinear):
                initial_currents[net] = v(0)
            else:
                initial_currents[net] = v
        return initial_currents
        

    def _default_circuit(self) -> str:
        """
        Create SPICE netlist of this simulation setup.
        """

        # Create a list of include files.
        setup_statements_string = ""
        if self.setup_statements is not None:
            # Load include files.
            setup_statements_string = "\n".join(self.setup_statements)

        # If the ground net is not called GND, then it must be connected to GND or 0.
        # This seems to be a must for SPICE simulators.
        # See: https://codeberg.org/librecell/lctime/issues/12#issuecomment-2173526
        ground_connection = ""
        if self.ground_net not in ["GND", "gnd", "0"]:
            logger.debug(f"Connect ground net `{self.ground_net}` to simulators ground `0`.")
            ground_connection = f"* Connect ground net to GND\n" \
            f"vgnd {self.ground_net} 0 0"

        # Load capacitance statements.
        load_capacitance_statements = ""
        if self.output_load_capacitances is not None:
            assert isinstance(self.output_load_capacitances, dict)
            load_capacitance_statements = "\n".join(
                (
                    f"Cload_{net} {net} {self.ground_net} {load}"
                    for net, load in self.output_load_capacitances.items()
                )
            )

        # Generate SPICE statements of input sources.
        signal_source_statements = ""
        signal_source_statements += _generate_input_voltage_statements(self.ground_net, self.input_voltages)
        signal_source_statements += "\n"
        signal_source_statements += _generate_input_current_statements(self.ground_net, self.input_currents)

        sim_netlist = f"""* librecell {__name__}
.title {self.simulation_title}

.control
* Exit on error with return value 1
pre_set strict_errorhandling
.endc

.option TEMP={self.temperature}

{self._params_initial_counditions()}

{setup_statements_string}

{ground_connection}

* Output load capacitances.
{load_capacitance_statements}

* Input and supply sources.
{signal_source_statements}

Xcircuit_under_test {" ".join(self.cell_ports)} {self.cell_name}

{self._initial_conditions_stmts()}

.end
"""
        return sim_netlist

    def _initial_conditions_stmts(self):
        # TODO
        # Also all voltages of DC sources must be here if they are needed to compute the initial conditions.

        initial_voltages = [
            f"v({net})={{ic_{net.lower()}}}"
            for net, _ in self._initial_voltages().items()
        ]
        initial_currents = [] # TODO
        # initial_currents = [
        #     f"i(V{net})={{ic_i_{net.lower()}}}"
        #     for net, _ in self._initial_currents().items()
        # ]
        ics = initial_voltages + initial_currents

        return f'.ic {" ".join(ics)}'

    def _params_initial_counditions(self):
        """
        Define variables (parameters) for the values of initial conditions. 
        """
        params = []

        for net, v in self._initial_voltages().items():
            params.append(f".param ic_{net.lower()}={v}")

        for net, i in self._initial_currents().items():
            params.append(f".param ic_i_{net.lower()}={i}")

        return "\n".join(params)

    def _init_simulation(self):
        logger.debug(self._default_circuit())
        #self.simulator.remove_circuit() # Remove any existing circuit.
        self.simulator.load_circuit(self._default_circuit())

    def set_temperature(self, temp: float):
        self.temperature = temp
        self.simulator.cmd(f"option TEMP={temp}")

    def set_breakpoint(self, breakpoint: str):
        assert breakpoint.startswith("stop")
        self.simulator.cmd(breakpoint.lower())

    # TODO remove
    # def delete_all_breakpoints(self):
    #     """
    #     Delete breakpoints.
    #     """
    #     self.simulator.drop_stdout()

    #     # Ask for current breakpoints.
    #     self.simulator.cmd("status")

    #     nums = [] # Breakpoint numbers to be deleted.
        
    #     read_timeout = 0.1
    #     # Get current breakpoints.
    #     while True:
    #         line = self.simulator.readline(timeout=read_timeout)
    #         read_timeout = 0.001 # Use shorter timeout after the first line.
    #         if line is None:
    #             break
    #         # Parse response.
    #         parts = line.split() # Split at whitespace
    #         if len(parts) > 1 and parts[1] == "stop":
    #             num = int(parts[0])
    #             nums.append(num)

    #     logger.info(f"num breakpoints: {len(nums)}")
    #     for num in nums:
    #         self.simulator.cmd(f"delete {num}")

    def set_capacitance(self, name: str, value: float):
        name = name.lower()
        assert name.startswith("c"), "capacitor name must start with C"
        self.simulator.cmd(f"alter {name} = {value}")
        
    def set_capacitor_voltage(self, name: str, value: float):
        name = name.lower()
        assert name.startswith("c"), "capacitor name must start with C"
        self.simulator.cmd(f"alter {name} ic = {value}")

    def set_resistance(self, name: str, value: float):
        name = name.lower()
        assert name.startswith("r"), "resistor name must start with R"
        self.simulator.cmd(f"alter {name} = {value}")

    def set_source_dc(self, name: str, value: float):
        name = name.lower()
        assert name.startswith("v") or name.startswith("i"), "source name must start with i or v"
        self.simulator.cmd(f"alter {name} dc = {value}")

    def set_source_pwl(self, name: str, pwl: PieceWiseLinear):
        name = name.lower()
        assert name.startswith("v") or name.startswith("i"), "source name must start with i or v"
        assert isinstance(pwl, PieceWiseLinear)
        self.simulator.cmd(f"alter {name} pwl = [ {pwl.to_spice_pwl_string()} ]")

    def set_source(self, name: str, source: Union[float, PieceWiseLinear]):
        if isinstance(source, PieceWiseLinear):
            self.set_source_dc(name, 0.0)
            self.set_source_pwl(name, source)
        else:
            self.set_source_dc(name, source)

    def set_initial_voltage(self, net: str, value: float):
        assert isinstance(value, float) or isinstance(value, int)
        self.simulator.cmd(f"alterparam ic_{net.lower()}={value}")

    def set_initial_current(self, component: str, value: float):
        assert isinstance(value, float) or isinstance(value, int)
        self.simulator.cmd(f"alterparam ic_i_{component.lower()}={value}")

    def tran(self, t_step: float, t_stop: float, t_start=None,
            output_voltages: List[str] = [], 
            output_currents: List[str] = [],
             ) -> Tuple[np.ndarray, Dict[str, np.ndarray], Dict[str, np.ndarray]]:
        """
        Run transient simulation.

        Return number of datapoints.
        """
        return self.simulator.tran(
            t_step=t_step, t_stop=t_stop, t_start=t_start,
            output_voltages=output_voltages, output_currents=output_currents
        )

    def reset(self):
        """
        Reset parameters, clear all breakpoints.
        """
        self.simulator.reset()
        self.simulator.cmd("delete all") # Delete all breakpoints.
        self.simulator.cmd("destroy all") # Remove previous datasets.

def _generate_input_voltage_statements(ground_net: str, input_voltages: Dict[str, Union[float, PieceWiseLinear]]) -> str:
    # Generate SPICE statements for input voltages.
    source_statements = []
    for net, wave in input_voltages.items():
        assert isinstance(wave, float) or isinstance(wave, PieceWiseLinear)
        source_statements.extend([
            __create_voltage_source_statement(ground_net, f"{net}", wave), # Ideal voltage source
        ])
    return "\n".join(source_statements)

def _generate_input_current_statements(ground_net: str, input_currents: Dict[str, Union[float, PieceWiseLinear]]) -> str:
    # Generate SPICE statements for input currents.
    source_statements = []
    for net, wave in input_currents.items():
        assert isinstance(wave, float) or isinstance(wave, PieceWiseLinear)
        source_statements.extend([
            __create_current_source_statement(ground_net, f"{net}", wave), # Ideal current source
        ])
    return "\n".join(source_statements)

def __create_voltage_source_statement(ground_net: str, net: str, voltage: Union[float, PieceWiseLinear]) -> str:
    """
    Create a SPICE statement for a voltage source driving the 'net' with a voltage.
    The voltage can be either a static value or a `PieceWiseLinear` function.
    """
    if isinstance(voltage, PieceWiseLinear):
        return f"V{net} {net} {ground_net} PWL({voltage.to_spice_pwl_string()}) DC=0"
    elif isinstance(voltage, float):
        return f"V{net} {net} {ground_net} {voltage}"
    else:
        assert False, "`voltage` must be either a float or {}".format(PieceWiseLinear)

def __create_current_source_statement(ground_net: str, net: str, current: Union[float, PieceWiseLinear]) -> str:
    """
    Create a SPICE statement for a current source driving the 'net' with a current.
    The current can be either a static value or a `PieceWiseLinear` function.
    """
    if isinstance(current, PieceWiseLinear):
        return f"I{net} {ground_net} {net} PWL({current.to_spice_pwl_string()}) DC=0"
    elif isinstance(current, float):
        return f"I{net} {ground_net} {net} {current}"
    else:
        assert False, "`current` must be either a float or {}".format(PieceWiseLinear)

class Simulator:
    """
    Abstract interface of SPICE circuit simulators.
    """

    def __init__(self):
        pass

    def name(self) -> str:
        """
        Get name of the simulator.
        """
        raise NotImplemented()

    def version_string(self) -> str:
        """
        Get version string of the simulator.
        """
        raise NotImplemented()

    def source(self, file: str):
        """
        Load a netlist from a file.
        """
        raise NotImplemented()

    def reset(self):
        raise NotImplemented()

    def tran(self, t_step, t_stop, t_start=None,
            output_voltages: List[str] = [], 
            output_currents: List[str] = [],
             ) -> Tuple[np.ndarray, Dict[str, np.ndarray], Dict[str, np.ndarray]]:
        """

        :param t_step:
        :param t_stop:
        :param t_start:
        :param output_voltages: Names of voltages which should be measured and returned
        :param output_currents:
        :return: A tuple of `(time, voltages, currents)`. Voltages and currents are stored in a dict by their name.
        """
        raise NotImplemented()

    def remove_circuit(self):
        """
        Remove current circuit.
        """
        raise NotImplemented()

    def load_circuit(self, circuit: str):
        """
        Load a circuit over stdin.
        :param circuit:
        :return:
        """
        raise NotImplemented()

class SimulatorSubprocess(Simulator):

    def __init__(self, executable: str, args: List[str] = []):
        self.executable = executable
        self.args = args

        self.proc: subprocess.Popen[str] = None
        "process handle."

        self._from_stdout = queue.Queue()
        self._from_stderr = queue.Queue()

        self._stdout_reader_thread = None
        self._stderr_reader_thread = None

    def _start(self):
        self.proc = subprocess.Popen(
            [self.executable] + self.args,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True
        )

        def read_stdout():
            while self.proc:
                line = self.proc.stdout.readline()
                if line is None:
                    break
                line = line.strip()
                if not line:  # Skip empty lines.
                    continue
                # print("stdout:", line)
                self._from_stdout.put(line)

        def read_stderr():
            while self.proc:
                line = self.proc.stderr.readline()
                if line is None:
                    break
                line = line.strip()
                if not line:  # Skip empty lines.
                    continue
                # print("stderr:", line)
                # Forward error messages to the logger.
                if 'Error' in line:
                    self.logger.error(line)
                elif 'Warning' in line:
                    self.logger.warning(line)
                self._from_stderr.put(line)

        self._stdout_reader_thread = threading.Thread(target=read_stdout, daemon=True)
        self._stderr_reader_thread = threading.Thread(target=read_stderr, daemon=True)
        self._stdout_reader_thread.start()
        self._stderr_reader_thread.start()

    def _stop(self):
        if self.proc:
            self.proc.kill()
            self.proc = None

    def readline(self, timeout: Optional[int] = 0.1) -> str:
        try:
            return self._from_stdout.get(block=True, timeout=timeout)
        except queue.Empty:
            return None

    def drop_stdout(self):
        """
        Delete the stdout buffer.
        """
        while not self._from_stdout.empty():
            self._from_stdout.get_nowait()

    def drop_stderr(self):
        """
        Delete the stderr buffer.
        """
        while not self._from_stdout.empty():
            self._from_stdout.get_nowait()

    def readline_err(self, timeout: float = 0.1) -> str:
        """
        Read stderr.
        Returns `None` if there's nothing on stderr within the timeout.
        """
        try:
            return self._from_stderr.get(block=True, timeout=timeout)
        except queue.Empty:
            return None

    def _write(self, data: str):
        self.proc.stdin.write(data)

    def _flush(self):
        self.proc.stdin.flush()

def test_ngspice_file_based():
    import math

    sim = NgSpiceFileBased()

    vdd = 1
    vcap_init = 0
    sim.load_circuit(f"""
V1 vout 0 {vdd}
R1 vout vcap 1000
C1 vcap 0 1e-6

.ic v(vcap)={vcap_init}
""")

    for _ in range(3): # Do simulation multiple times to check if 'reset' works.
        t_stop = 1
        time, voltages, currents = sim.tran(
                    t_step=1e-4, t_stop=t_stop,
                    output_voltages=['vcap', 'vout'],
                    output_currents=['v1']
                )

        assert abs(time[-1] - t_stop) < 1e-6
        assert np.all(np.isclose(voltages['vout'], vdd))
        assert abs(voltages['vcap'][0] - vcap_init) < 1e-3
        assert abs(voltages['vcap'][-1] - vdd) < 1e-3

        sim.reset()

class NgSpiceFileBased(SimulatorSubprocess):
    """
    Interface to non-interactive ngspice.
    Writes the simulation script into a file, calls ngspice, then reads the results.
    """

    def __init__(self,
                 ngspice_executable: str = None,
                 workdir: str = None,
                 logger: logging.Logger = None):
        self.logger = logger
        if logger is None:
            self.logger = logging.getLogger(__name__)
        executable = 'ngspice' if ngspice_executable is None else ngspice_executable

        super().__init__(executable, ["-p"])

        self._workdir = workdir
        if workdir is None:
            self._workdir = tempfile.mkdtemp(prefix="lctime")

        self._commands = []
        self._circuit: str = ""

    def name(self) -> str:
        return "ngspice-file-based"

    def version_string(self) -> str:
        """
        Get ngspice version string.
        Returns something like "ngspice-39" or "ngspice-44+".
        Raises an exception if ngspice executable cannot be found.
        """

        output = subprocess.check_output([self.executable, "--version"]).decode("utf-8")
        lines = output.split("\n")

        for l in lines:
            if "ngspice-" in l:
                parts = l.split()
                if len(parts) > 2:
                    return parts[1]

        return "ngspice-unknown"

    def __exit__(self, *args):
        pass

    def __enter__(self):
        return self
    
    def cmd(self, cmd: str):
        self._commands.append(cmd)

    def source(self, file: str):
        """
        Load a netlist.
        """
        self.cmd(f"source {file}")

    def reset(self):
        #self._commands.clear()
        def take_cmd(c: str) -> bool:
            return c.startswith("alter")
        self._commands = [c for c in self._commands if take_cmd(c)]
        self.cmd('reset')
        
    def remove_circuit(self):
        self._commands = []
        self._circuit = ""
    
    def load_circuit(self, circuit: str):
        self._commands = []
        self._circuit = circuit

    def _write_sim_file(self, path: str):

        control = "\n".join(self._commands)
        

        spice_script = f"""
.title lctime simulation

{self._circuit}

.control
{control}
.endc

.end
        """

        with open(path, "w") as f:
            f.write(spice_script)

    def _run_simulation(self, output_voltages: List[str], output_currents: List[str]):
        simfile = os.path.join(self._workdir, "simulation.sp")
        sim_output_file = os.path.join(self._workdir, "simulation_out.txt")
        
        output_voltage_stmts = " ".join(f"v({v})" for v in output_voltages)
        output_current_stmts = " ".join(f"i({v})" for v in output_currents)
        
        self.cmd("set filetype=ascii")
        self.cmd("set wr_vecnames") # Enable output of vector names in the first line
        self.cmd(f"wrdata {sim_output_file} {output_voltage_stmts} {output_current_stmts}")
        self.cmd("exit")

        self._write_sim_file(simfile)

        result = subprocess.run(
            [self.executable, simfile],
            capture_output=True
        )

        if result.returncode != 0:
            raise Exception(f"ngspice failed: {result.stderr}")
        
    def _get_data(self, output_voltages: List[str] = [], output_currents: List[str] = []):
        sim_output_file = os.path.join(self._workdir, "simulation_out.txt")
        #data = np.genfromtxt(sim_output_file, skip_header=1, ndmin=2)
        data = np.loadtxt(sim_output_file, skiprows=1, ndmin=2)

        time = data[:,0]

        # Extract signals (strip away time)
        data = data[:,1::2]
        voltages = data[:, :len(output_voltages)].T
        currents = data[:, len(output_voltages):].T
        assert len(currents) == len(output_currents), "mismatch in number of data columns"
        
        voltages = {name: v for name, v in  zip(output_voltages, voltages)}
        currents = {name: i for name, i in  zip(output_currents, currents)}

        return time, voltages, currents

    def tran(self,
            t_step: float,
            t_stop: float, 
            output_voltages: List[str] = [], 
            output_currents: List[str] = [],
            t_start: float=None,
            uic: str = ""
            ) -> Tuple[np.ndarray, Dict[str, np.ndarray], Dict[str, np.ndarray]]:

        self.cmd(f'tran {t_step} {t_stop} {uic}')
        self._run_simulation(output_voltages, output_currents)
        
        return self._get_data(output_voltages, output_currents)

class NgSpiceInteractive(SimulatorSubprocess):
    """
    Interface to ngspice as an interactive subprocess.
    """

    def __init__(self,
                 ngspice_executable: str = None,
                 logger: logging.Logger = None):
        self.logger = logger
        if logger is None:
            self.logger = logging.getLogger(__name__)
        executable = 'ngspice' if ngspice_executable is None else ngspice_executable

        super().__init__(executable, ["-p"])

        self._circuit_is_loaded = False

    def name(self) -> str:
        return "ngspice-interactive"

    def version_string(self) -> str:
        """
        Get ngspice version string.
        Returns something like "ngspice-39" or "ngspice-44+".
        Raises an exception if ngspice executable cannot be found.
        """

        output = subprocess.check_output([self.executable, "--version"]).decode("utf-8")
        lines = output.split("\n")

        for l in lines:
            if "ngspice-" in l:
                parts = l.split()
                if len(parts) > 2:
                    return parts[1]

        return "ngspice-unknown"

    def __reboot(self):
        """
        Restart ngspice process. Used as a workaraound for memory leaks in ngspice.
        """
        self._stop()
        self._start();

    def __exit__(self, *args):
        self._stop()

    def __enter__(self):
        self._start()
        return self

    def cmd(self, cmd: str):
        self.logger.debug(f"ngspice> {cmd}")
        self._write(cmd)
        self._write('\n')
        self._flush()
        self.__check_err(0.001) # TODO remove?

    def source(self, file: str):
        """
        Load a netlist.
        """
        self.cmd(f"source {file}")

    def reset(self):
        self.cmd('reset')

    def _tran(self, t_step, t_stop, t_start=None, uic='') -> int:
        """

        :param t_step:
        :param t_stop:
        :param t_start:
        :param uic:
        :return: Return number of data points.
        """
        self._flush()
        self.cmd(f'tran {t_step} {t_stop} {uic}')
        self.cmd(f'print length(time)')  # Should print something like: length(time) = 1.234e+03

        # Find number of data points.
        num_rows = 0
        while True:
            line = self.readline(timeout=None)
            assert line is not None
            if line.startswith('length(time)'):
                _, num_rows_str = line.split('=', 2)
                num_rows = int(float(num_rows_str))
                break
            else:
                logger.debug(f"ngspice: {line}")

        assert num_rows > 0
        return num_rows

    def remove_circuit(self):
        """
        Remove current circuit.
        """
        self.cmd("remcirc")
        self._circuit_is_loaded = False

    def load_circuit(self, circuit: str):
        """
        Load a circuit over stdin.
        :param circuit:
        :return:
        """
        
        if self._circuit_is_loaded:
            self.remove_circuit()
            
        lines = circuit.split('\n')
        for line in lines:
            line = line.strip()
            if line:
                self._write("circbyline ")
                self._write(line)
                self._write('\n')
                self._flush()
        self.__check_err(timeout=0)

        self._circuit_is_loaded = True

    def tran(self,
            t_step: float,
            t_stop: float, 
            output_voltages: List[str] = [], 
            output_currents: List[str] = [],
            t_start: float=None,
            ) -> Tuple[np.ndarray, Dict[str, np.ndarray], Dict[str, np.ndarray]]:
        num_rows = self._tran(t_step=t_step, t_stop=t_stop, t_start=t_start)
        
        return self._get_data(num_rows, output_voltages, output_currents)

    def _get_data(self, 
                 num_rows: int, 
                 output_voltages: List[str] = [], 
                 output_currents: List[str] = []
             ) -> Tuple[np.ndarray, Dict[str, np.ndarray], Dict[str, np.ndarray]]:
        """
        Get data points.
        :param num_rows: number of data points created by `self.tran()`
        """

        sim_data = self._get_data_raw(
            num_rows, output_voltages, output_currents
        )
        
        if sim_data.ndim != 2:
            logger.error("Simulation failed. No data was written to the output file.")
            if debug:
                pass
                # TODO logger.error(f"ngspice: {stderr}")
            assert False, "Simulation failed. No data was written to the output file."

        # Extract data from the numpy array.
        time = sim_data[:, 0]
        index = 1

        # Put voltages into a dict.
        voltages = dict()
        for v_out in output_voltages:
            voltages[v_out] = sim_data[:, index]
            index = index + 1

        # Put currents into a dict.
        currents = dict()
        for i_out in output_currents:
            currents[i_out] = sim_data[:, index]
            index = index + 1

        return time, voltages, currents

    def _get_data_raw(self, num_rows: int, voltages: List[str] = None, currents: List[str] = None) -> np.ndarray:
        # TODO use wrdata into a pipe (os.mkfifo()).
        self.drop_stdout()

        signals = []
        if voltages is not None:
            signals.extend((f"v({v})" for v in voltages))

        if currents is not None:
            signals.extend((f"i({i})" for i in currents))

        signals_str = " ".join(signals)

        self.cmd(f"set width={(1 + len(signals)) * 40}")
        self.cmd(f"print {signals_str}")

        rows = []
        i = 0
        while True:
            line = self.readline(timeout=1)
            assert line is not None
            if line.startswith(str(i)):
                i = i + 1
                arr = np.fromstring(line, dtype=float, sep=' ')
                rows.append(arr)
            if i == num_rows:
                break

        data = np.array(rows)
        # Remove index.
        data = data[:, 1:]

        return data

    def __check_err(self, timeout: float = 0.1):
        """
        Raise an exception on an error.
        """

        while True:
            err = self.readline_err(timeout=timeout)
            if err is None:
                break
            if err.startswith("Error"):
                raise Exception(f"ngspice error: {err}")




# SPDX-FileCopyrightText: 2025 Thomas Kramer <code@tkramer.ch>
# SPDX-License-Identifier: AGPL-3.0-or-later

from .util import CellConfig, CharacterizationConfig
from .ngspice_subprocess import run_simulation
from ..lccommon.net_util import get_subcircuit_ports

from typing import List
import os
import logging

logger = logging.getLogger(__name__)

def flatten_circuit(
    cell_name: str,
    spice_netlist_file: str,
    setup_statements: List[str],
    workingdir: str,
) -> str:
    """
    Flatten a SPICE circuit using ngspice.
    This is used to resolve transistors which are wrapped in subcircuits.

    Returns a flatened SPICE subcircuit.
    """

    logger.debug("flatten netlist with ngspice")
    
    # Create a list of include files.
    setup_statements = setup_statements + [f".include {spice_netlist_file}"]
    
    ports = get_subcircuit_ports(spice_netlist_file, cell_name)

    # Load include files.
    for setup in setup_statements:
        logger.debug(f"Setup statement: {setup}")
    setup_statements_string = "\n".join(setup_statements)

    # Simulation script file path.
    file_name = f"lctime_flatten_{cell_name}"
    sim_file = os.path.join(workingdir, f"{file_name}.sp")

    # Create ngspice simulation script.
    sim_netlist = f"""* librecell {__name__}
.title Flatten circuit

.control
* Exit on error with return value 1
pre_set strict_errorhandling
.endc

{setup_statements_string}

Xcell {" ".join(ports)} {cell_name}

.control

set filetype=ascii
set wr_vecnames

listing r
exit
.endc

.end
"""

    # Dump netlist.
    logger.debug(sim_netlist)

    # Dump simulation script to the file.
    # logger.debug(f"Write simulation netlist: {sim_file}")
    open(sim_file, "w").write(sim_netlist)

    # Run simulation.
    stdout, stderr = run_simulation(sim_file)

    flattened = [f'.subckt {cell_name} {" ".join(ports)}']
    # Fetch flat netlist.
    for line in stdout.split("\n"):
        if line.startswith("m.xcell"):
            # Drop parameters of transistor.
            t = line.split()
            flattened.append(" ".join(t[:6]))
        if line.startswith("r.xcell"):
            flattened.append(line)
    flattened.append(".ends")


    return "\n".join(flattened)




# Copyright (c) 2019-2020 Thomas Kramer.
# SPDX-FileCopyrightText: 2022 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
Characterization functions for combinatorial cells.
"""

from typing import List, Dict, Callable, Optional

from itertools import product
import os

from .util import *
from .piece_wise_linear import *
from .ngspice_subprocess import CellSimulation, NgSpiceInteractive
from ..lccommon.net_util import get_subcircuit_ports

import logging

logger = logging.getLogger(__name__)


def characterize_comb_cell(
        simulation: CellSimulation,
        input_pins: List[str],
        output_pin: str,
        related_pin: str,
        output_functions: Dict[str, Callable],

        total_output_net_capacitance: np.ndarray,
        input_net_transition: np.ndarray,

        cell_conf: CellConfig,

        constant_inputs: Dict[str, bool] = dict(),
        initial_node_voltages: Dict[str, float] = None,
        static_input_patterns: Optional[List[Dict[str, bool]]] = None

) -> Dict[str, np.ndarray]:
    """
    Calculate the NLDM timing table of a cell for a given timing arc.
    :param simulation: simulation harness
    :param constant_inputs: A truth-assignment for input pins that are static. Used for tri-state enable pins.
    :param cell_conf: Parameters and configuration for the characterization.
    :param input_net_transition: Transition times of input signals in seconds.
    :param total_output_net_capacitance: Load capacitance in Farads.
    :param input_pins: List of input pins.
    :param output_pin: The output pin of the timing arc.
    :param related_pin: The input pin of the timing arc.
    :param output_functions: A dict mapping output pin names to corresponding boolean functions.
    :param initial_conditions: Initial node voltages as a dict.
    :param static_input_patterns: Use this assignments of the static input pins for characterization. Characterize for each assignment. If this is not supplied, characterization is run for all combinations.

    :return: Returns the NLDM timing tables wrapped in a dict:
    {'cell_rise': 2d-np.ndarray, 'cell_fall': 2d-np.ndarray, ... }
    """
    assert isinstance(simulation, CellSimulation)
    assert isinstance(cell_conf, CellConfig)
    assert isinstance(cell_conf.global_conf, CharacterizationConfig)
    cfg = cell_conf.global_conf

    inputs_inverted = cell_conf.complementary_pins.values()
    assert related_pin not in inputs_inverted, f"Active pin '{related_pin}' must not be an inverted pin of a differential pair."
    input_pins_non_inverted = [p for p in input_pins if p not in inputs_inverted]
    related_pin_inverted = cell_conf.complementary_pins.get(related_pin)

    vdd = cfg.supply_voltage

    # TODO set reasonable upper bound for simulation time
    # Maximum simulation time.
    time_max = cfg.time_step * 1e5

    # Find function to summarize different timing arcs.
    # TODO: Make this directly parametrizable by caller.
    reduction_function = {
        CalcMode.WORST: max,
        CalcMode.BEST: min,
        CalcMode.TYPICAL: np.mean
    }[cfg.timing_corner]
    
    # Initial node voltages.
    if initial_node_voltages is None:
        initial_node_voltages = dict()
    else:
        initial_node_voltages = initial_node_voltages.copy()

    # Get all input nets that are not toggled during a simulation run.
    static_input_nets = [i for i in input_pins_non_inverted if i != related_pin and i not in constant_inputs]

    if static_input_patterns is None:
        static_input_patterns = [{}] # Add a pattern which does not constrain any input and therefore leads to all possible combinations.
    else:
        # Sanity test on input.
        assert isinstance(static_input_patterns, List)
        for pat in static_input_patterns:
            for var in pat.keys():
                assert var in static_input_nets, f"input name in static_input_patterns is not in the list of static inputs: {var}"

    # Get a list of all input combinations that will be used for measuring conditional timing arcs.
    # Generate all possible input combinations at static inputs.
    static_inputs = []
    for pattern in static_input_patterns:
        assert isinstance(pattern, Dict)

        possible_assignments = [] # List of possible assignemnts for each static input.
        # For each static input, get the possible assignments allowed by this pattern.
        for i in static_input_nets:
            possible_assignment = [0, 1] # If the pattern does not specify an assignment, both 0 and 1 are possible.
            if i in pattern:
                possible_assignment = [1 if pattern[i] == True else 0]
            possible_assignments.append(possible_assignment)

        static_inputs.extend(
            list(product(*possible_assignments))
        )
        
    # Get boolean function of output pin.
    assert output_pin in output_functions, \
        "Boolean function not defined for output pin '{}'".format(output_pin)
    output_function = output_functions[output_pin]

    def f(input_transition_time: float, output_cap: float):
        """
        Evaluate cell timing at a single input-transition-time/output-capacitance point.
        :param input_transition_time: Transition time (slew) of the active input signal.
        :param output_cap: Load capacitance at the output.
        :return:
        """

        # Empty results.
        rise_transition_durations = []
        fall_transition_durations = []
        rise_delays = []
        fall_delays = []

        rise_powers = []
        fall_powers = []

        for static_input in static_inputs:

            # Check if the output is controllable with this static input.
            bool_inputs = {net: value > 0 for net, value in zip(static_input_nets, static_input)}

            bool_inputs[related_pin] = False
            output_when_false = output_function(**bool_inputs)
            bool_inputs[related_pin] = True
            output_when_true = output_function(**bool_inputs)

            if output_when_false == output_when_true:
                # The output will not change when this input is changed.
                # Simulation of this conditional arc can be skipped.
                logger.debug("Simulation skipped for conditional arc (output does not toggle): {}".format(bool_inputs))
                continue

            # Create DC values to be applied to the input pins.
            input_voltages = _create_static_input_voltages(
                cell_conf=cell_conf,
                static_input_nets=static_input_nets,
                static_input=static_input,
                constant_inputs=constant_inputs,
            )

            logger.debug("Static input voltages: {}".format(input_voltages))

            for input_rising in [True, False]:
                simulation.reset() # Rewind the simulation.

                # Simulation script file path.
                file_name = f"lctime_combinational_" \
                            f"slew={input_transition_time}_" \
                            f"load={output_cap}" \
                            f"{''.join((f'{net}={v}' for net, v in input_voltages.items() if isinstance(v, float)))}_" \
                            f"{'rising' if input_rising else 'falling'}"

                # File for debug plot of the waveforms.
                sim_plot_file = os.path.join(cfg.workingdir, f"{file_name}_plot.svg")

                bool_inputs[related_pin] = input_rising
                expected_output = output_function(**bool_inputs)
                initial_output_voltage = 0.0 if expected_output else vdd
                initial_node_voltages[output_pin] = initial_output_voltage

                # Get stimulus signal for related pin.
                input_wave = StepWave(start_time=0, polarity=input_rising,
                                      rise_threshold=0,
                                      fall_threshold=1,
                                      transition_time=input_transition_time)
                input_wave.y = input_wave.y * vdd
                input_voltages[related_pin] = input_wave
                # Get stimulus signal for the inverted pin (if any).
                if related_pin_inverted:
                    input_wave_inverted = vdd - input_wave
                    input_voltages[related_pin_inverted] = input_wave_inverted


                # Set initial conditions
                for net, value in input_voltages.items():
                    if isinstance(value, float):
                        simulation.set_initial_voltage(net, value)
                    elif isinstance(value, PieceWiseLinear):
                        simulation.set_initial_voltage(net, float(value(0)) )
                        
                # # Set initial node voltages.
                # for pin, voltage in initial_node_voltages.items():
                #     if pin == output_pin: # TODO
                #         #simulation.set_capacitor_voltage(f"Cload_{pin}", voltage)
                #         pass
                #     simulation.set_initial_voltage(pin, float(voltage))
                simulation.set_initial_voltage(output_pin, initial_output_voltage)

                # Need to call `reset` after altering parameters (see ngspice user manual, section 13.5.5).
                simulation.simulator.cmd("reset")
                
                    
                # Get the breakpoint condition.
                if expected_output:
                    simulation.set_breakpoint(f"stop when v({output_pin}) > {vdd * 0.99}")
                else:
                    simulation.set_breakpoint(f"stop when v({output_pin}) < {vdd * 0.01}")

                # Set input voltages.
                for net, value in input_voltages.items():
                    src = f"V{net}"
                    simulation.set_source(src, value)

                # Set output load.
                simulation.set_capacitance(f"Cload_{output_pin}", output_cap)

                if cfg.debug:
                    # Dump current simulation setup which is loaded in ngspice.
                    simulation.simulator.cmd("listing")
                    
                # Run simulation
                time, voltages, currents = simulation.tran(
                    t_step=cfg.time_step, t_stop=time_max,
                    output_voltages=[related_pin, output_pin],
                    output_currents=[f"V{cell_conf.supply_net}", f"V{related_pin}"],
                )
      
                # Retrieve data.
                supply_current = - currents[f"V{cell_conf.supply_net}"]
                gate_current = - currents[f"V{related_pin}"]
                
                input_voltage = voltages[related_pin]
                output_voltage = voltages[output_pin]

                if cfg.debug_plots:
                    logger.debug("Create plot of waveforms: {}".format(sim_plot_file))
                    import matplotlib
                    matplotlib.use('Agg')
                    import matplotlib.pyplot as plt
                    plt.close()
                    plt.title(f"")
                    plt.plot(time, input_voltage, 'x-', label='input voltage')
                    plt.plot(time, output_voltage, label='output voltage')
                    plt.plot(time, supply_current, label='supply current')
                    plt.legend()
                    plt.savefig(sim_plot_file)
                    plt.close()

                # TODO: What unit does rise_power/fall_power have in liberty files???
                # Is it really power or energy?
                dt = time[-1] - time[0]
                assert dt > 0.
                gate_energy = np.mean(gate_current * input_voltage) * dt
                supply_energy = np.mean(supply_current * vdd) * dt
                logger.debug(f"switching energy (supply): {supply_energy}")
                logger.debug(f"switching energy (gate): {gate_energy}")

                switching_energy = gate_energy + supply_energy
                if switching_energy < 0:
                    logger.warn("Negative switching energy: {switching_energy}")
                
                v_thresh = 0.5 * vdd

                # Get input signal before switching and after.
                input_a = input_voltage[0] > v_thresh
                input_b = input_voltage[-1] > v_thresh
                assert input_a != input_b

                # Get output signal before switching and after.
                output_a = output_voltage[0] > v_thresh
                output_b = output_voltage[-1] > v_thresh

                # There should be an edge in the output signal.
                # Because the input signals have been chosen that way.
                assert output_a != output_b, "Supplied boolean function and simulation are inconsistent."

                output_rising = output_b

                # Normalize input/output such that both have a rising edge.
                input_voltage = input_voltage if input_rising else vdd - input_voltage
                output_voltage = output_voltage if output_rising else vdd - output_voltage
                # Normalize to range [0, ..., 1]
                input_voltage /= vdd
                output_voltage /= vdd

                # Check if signals are already stabilized after one `period`.
                # assert abs(input_voltage[0]) < 0.01, "Input signal not yet stable at start."
                # assert abs(1 - input_voltage[-1]) < 0.01, "Input signal not yet stable at end."
                if output_rising:
                    output_threshold = cfg.trip_points.output_threshold_rise
                else:
                    output_threshold = cfg.trip_points.output_threshold_fall
                assert abs(output_voltage[0]) <= output_threshold, "Output signal not yet stable at start."
                assert abs(1 - output_voltage[-1]) <= output_threshold, "Output signal not yet stable at end."

                # Calculate the output slew time: the time the output signal takes to change from
                # `slew_lower_threshold` to `slew_upper_threshold`.
                output_transition_duration = get_slew_time(time, output_voltage, trip_points=cfg.trip_points)

                # Calculate delay from the moment the input signal crosses `input_threshold` to the moment the output
                # signal crosses `output_threshold`.
                cell_delay = get_input_to_output_delay(time, input_voltage, output_voltage,
                                                       trip_points=cfg.trip_points)

                if output_rising:
                    rise_delays.append(cell_delay)
                    rise_transition_durations.append(output_transition_duration)
                else:
                    fall_delays.append(cell_delay)
                    fall_transition_durations.append(output_transition_duration)

                if input_rising:
                    rise_powers.append(switching_energy)
                else:
                    fall_powers.append(switching_energy)

        return (reduction_function(np.array(rise_delays)),
                reduction_function(np.array(fall_delays)),
                reduction_function(np.array(rise_transition_durations)),
                reduction_function(np.array(fall_transition_durations)),
                reduction_function(np.array(rise_powers)),
                reduction_function(np.array(fall_powers)),
                )

    f_vec = np.vectorize(f, cache=True)

    xx, yy = np.meshgrid(input_net_transition, total_output_net_capacitance)

    # Evaluate timing on the grid.
    cell_rise, cell_fall, rise_transition, fall_transition, rise_power, fall_power = f_vec(xx, yy)

    # Return the tables by liberty naming scheme.
    return {
        'total_output_net_capacitance': total_output_net_capacitance,
        'input_net_transition': input_net_transition,
        'cell_rise': cell_rise,
        'cell_fall': cell_fall,
        'rise_transition': rise_transition,
        'fall_transition': fall_transition,
        'rise_power': rise_power,
        'fall_power': fall_power
    }


def _create_static_input_voltages(
        cell_conf: CellConfig, 
        static_input_nets: List[str], 
        static_input: List[float],
        constant_inputs: Dict[str, bool]
    ) -> Dict[str, Union[float, PieceWiseLinear]]:
    """
    For each cell input pin create either a DC voltage or a piece-wise linear
    waveaform.
    """
    assert isinstance(cell_conf, CellConfig)
    assert len(static_input_nets) == len(static_input)
    assert isinstance(constant_inputs, dict)
    cfg = cell_conf.global_conf
    
    # Get voltages at static inputs.
    input_voltages = {net: cfg.supply_voltage * value for net, value in zip(static_input_nets, static_input)}
    # Add supply voltage.
    input_voltages[cell_conf.supply_net] = cfg.supply_voltage
    
    # Add nwell voltage
    if cell_conf.nwell_pin is not None:
        input_voltages[cell_conf.nwell_pin] = cfg.supply_voltage
        
    # Add pwell voltage
    if cell_conf.pwell_pin is not None:
        input_voltages[cell_conf.pwell_pin] = 0.0
        
    # Add input voltages for inverted inputs of differential pairs.
    for p in static_input_nets:
        inv = cell_conf.complementary_pins.get(p)
        if inv is not None:
            assert inv not in input_voltages
            # Add the inverted input voltage.
            input_voltages[inv] = cfg.supply_voltage - input_voltages[p]

    # Add constant voltages (tri-state enable pin).
    for pin, value in constant_inputs.items():
        input_voltages[pin] = cfg.supply_voltage if value else 0.0

    return input_voltages
    



# SPDX-FileCopyrightText: 2022 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

import tempfile
import subprocess
import numpy as np
import matplotlib.pyplot as plt
import os
import asyncio
import logging
import threading, queue
import time
from typing import List

from .ngspice_subprocess import NgSpiceInteractive

def test_version_number():

    with NgSpiceInteractive() as sim:
        version = sim.version_string()
        assert version is not None
        assert version.startswith("ngspice-")
        assert version == version.strip()

def test_simple_simulation():
    """Run a simple transient simulation in a ngspice subprocess and retreive the results from a file."""

    sim_file = tempfile.mktemp()
    sim_output_file = tempfile.mktemp()

    spice_simulation_netlist = f"""
*
.title Simple RC circuit.

R1 VDD Y 1k
C1 Y GND 1u ic=0V
Vsrc_vdd VDD GND PWL(0 0 1ms 0V 2ms 1V)

.control
*option abstol=10e-15
*option reltol=10e-11
set filetype=ascii
* Enable output of vector names in the first line.
set wr_vecnames
tran 1ms 10ms
wrdata {sim_output_file} v(VDD) v(Y)
exit
.endc

.end
"""
    print(f"Write simulation file: {sim_file}")
    open(sim_file, 'w').write(spice_simulation_netlist)

    # Run simulation.
    ret = subprocess.run(["ngspice", sim_file])
    print(f"Subprocess return value: {ret}")
    if ret.returncode != 0:
        print(f"ngspice simulation failed: {ret}")
    assert ret.returncode == 0

    print(f"Read output data: {sim_output_file}")
    data = np.loadtxt(sim_output_file, skiprows=1)  # Skip the header.

    a_time = data[:, 0]
    a = data[:, 1]

    y_time = data[:, 2]
    y = data[:, 3]

    assert all(a_time == y_time)

    plt.plot(a_time, a, 'x-')
    plt.plot(y_time, y, 'x-')
    # Disable for automatic tests: plt.show()

    os.remove(sim_output_file)
    os.remove(sim_file)


def test_interactive_subprocess():
    import subprocess
    p = subprocess.Popen(['ngspice', '-p'],
                         stdin=subprocess.PIPE,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE,
                         universal_newlines=True)

    p.stdin.write("a\n")
    p.stdin.flush()
    print(p.stderr.readline())

    p.stdin.write("quit\n")
    p.stdin.flush()
    print(p.stdout.readlines())
    print(p.stderr.readlines())


def test_ngspice_interactive_simple():

    with NgSpiceInteractive() as sim:

        sim.load_circuit(r"""
.title test RC circuit
C1 0 A 0.001
R1 0 A 100

.ic v(a)=2
.end
""")

        time, voltages, currents = sim.tran(
            t_step=0.001, t_stop=1,
            output_voltages=["A"]
        )
        voltage_a = voltages["A"]
        assert time[0] == 0.0

        assert abs(voltage_a[0]-2) < 1e-6
        assert voltage_a[-1] < 0.01, "voltage over capacitor should decay"
        
def test_ngspice_subprocess_class():

    spice_simulation_netlist = f"""
*
.title Simple RC circuit.

.param voltage_end = 1V

R1 VDD Y 1k
C1 Y GND 1u ic=0V
Vsrc_vdd VDD GND PWL(0 0 1ms 0V 2ms {{voltage_end}})

.end
    """

    with NgSpiceInteractive() as ns:

        # ns.source(sim_file)
        ns.load_circuit(spice_simulation_netlist)

        # ns.cmd(spice_simulation_netlist)
        # ns.cmd('set filetype=ascii')
        # ns.cmd('tran 1ms 1ms')
        time1, v1, i1 = ns.tran(t_step='1ms', t_stop='1ms',
             output_voltages=['VDD', 'Y'], output_currents=['vsrc_vdd']
         )
        assert v1['VDD'][0] == 0
        assert v1['VDD'][-1] == 0

        err = ns.readline_err()
        assert err == None, f"ngspice error: {err}"

        time, v2, i2 = ns.tran(t_step='1ms', t_stop='2ms', 
                output_voltages=['VDD', 'Y'], output_currents=['vsrc_vdd']
            )
        assert v2['VDD'][0] == 0
        assert v2['VDD'][-1] == 1
        # ns.cmd('wrdata /dev/stdout v(VDD) v(Y)')

        err = ns.readline_err()
        assert err == None, f"ngspice error: {err}"


        # Re-run the simulation with another wave form.
        ns.reset()
        # update the waveform
        ns.cmd("alter Vsrc_vdd pwl = [ 0 0 1ms 0V 2ms 2V ]")
        # Change capacitor value
        ns.cmd("alter C1 = 2u")
        #ns.cmd("alterparam voltage_end = 2V")
        time, v2, i2 = ns.tran(t_step='1ms', t_stop='2ms',
                        output_voltages=['VDD', 'Y'],
                        output_currents=['vsrc_vdd']
                    )
        err = ns.readline_err()
        assert err == None, f"ngspice error: {err}"
        assert v2["VDD"][0] == 0
        assert v2["VDD"][-1] == 2, "failed to alter the piece-wise linear voltage source"


#class NgSpiceAsync:
#    """
#    Asynchronous interface to ngspice as a subprocess.
#    """
#
#    def __init__(self, logger: logging.Logger = None):
#        if logger is None:
#            logger = logging.getLogger(__name__)
#        self.logger = logger
#        self.ngspice_cmd = "ngspice"
#        self.ngspice_args = ['-p']
#        "Run ngspice in interactive pipe mode."
#
#        self.proc = None
#        "ngspice process handle."
#
#    async def start(self):
#        self.proc = await asyncio.create_subprocess_exec(
#            self.ngspice_cmd,
#            *self.ngspice_args,
#            stdin=asyncio.subprocess.PIPE,
#            stdout=asyncio.subprocess.PIPE,
#            stderr=asyncio.subprocess.PIPE,
#        )
#
#    async def cmd(self, cmd: bytes):
#        if not cmd.endswith(b"\n"):
#            cmd += b"\n"
#        self.logger.debug(f"Run cmd: {cmd}")
#        print(f"Run cmd: {cmd}")
#        # cmd = cmd.encode('utf-8')
#        stdout, stderr = await self.proc.communicate(input=cmd)
#        self.logger.debug(f"stdout: {stdout}")
#        print(f"stdout: {stdout}")
#        self.logger.debug(f"stderr: {stderr}")
#        print(f"stderr: {stderr}")
#
#        print(f"return code: {self.proc.returncode}")
#
#
#def test_async_interactive_subprocess():
#
#    async def run():
#        
#        ns = NgSpiceAsync()
#        try:
#            async with asyncio.timeout(1):
#                await ns.start()
#                await ns.cmd(b"help")
#                await ns.cmd(b"help")
#                await ns.cmd(b"quit")
#        except Exception as e:
#            if ns.proc:
#                ns.proc.kill()
#            raise e
#
#    asyncio.run(run())
#
#
#def test_async_interactive_subprocess_raw():
#
#    async def run():
#        
#        proc = None
#        try:
#            async with asyncio.timeout(1):
#            
#                cmd = 'ngspice -p'
#                proc = await asyncio.create_subprocess_exec(
#                    'ngspice',
#                    '-p',
#                    stdin=asyncio.subprocess.PIPE,
#                    stdout=asyncio.subprocess.PIPE,
#                    stderr=asyncio.subprocess.PIPE,
#                )
#
#                stdout, stderr = await proc.communicate(input=b"help\n")
#                print(stdout)
#                print(stderr)
#                stdout, stderr = await proc.communicate(input=b"quit\n")
#
#                print(f'[{cmd!r} exited with {proc.returncode}]')
#                if stdout:
#                    print(f'[stdout]\n{stdout.decode()}')
#                if stderr:
#                    print(f'[stderr]\n{stderr.decode()}')
#        except Exception as e:
#            if proc:
#                proc.kill()
#            raise e
#
#    asyncio.run(run())
#



# SPDX-FileCopyrightText: 2024 Thomas Kramer <code@tkramer.ch>
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
Measurement of the input capacitance by driving the input pin with a constant current.
"""

from .util import *
from .piece_wise_linear import *
from .ngspice_subprocess import run_simulation
from ..lccommon.net_util import get_subcircuit_ports

import logging
from typing import Dict
from itertools import product
from scipy import interpolate

logger = logging.getLogger(__name__)


def characterize_leakage_power(
        input_pins: List[str],
        internal_state_nodes: List[str],
        cell_conf: CellConfig
) -> Dict[str, float]:
    pass



# SPDX-FileCopyrightText: 2019-2024 Thomas Kramer <code@tkramer.ch>
# SPDX-License-Identifier: AGPL-3.0-or-later

from .util import *
from ..cell_types import SingleEdgeDFF
from . import timing_sequential as seq
from ..liberty import util as liberty_util
from ..logic import seq_recognition
from ..characterization.ngspice_subprocess import NgSpiceInteractive, CellSimulation
from ..lccommon.net_util import get_subcircuit_ports

from liberty.types import *
import numpy as np
from sympy.logic import satisfiable

def characterize_flip_flop_output(
        simulation: CellSimulation,
        new_library: Group,
        new_cell_group: Group,
        conf: CharacterizationConfig,
        cell_type: SingleEdgeDFF,
        cell_conf: CellConfig,
        related_pin_transition: np.ndarray,
        input_transition_times: np.ndarray,
        output_capacitances: np.ndarray,
):
    assert isinstance(simulation, CellSimulation)
    assert isinstance(cell_type, SingleEdgeDFF)

    logger.info("Characterize single-edge triggered flip-flop.")

    if related_pin_transition is None:
        logger.error("'related-pin-transition' is not specified for the clock pin.")

    assert related_pin_transition is not None, "'related-pin-transition' is not specified for the clock pin."

    # Create or update the 'ff' group.
    ff_groups = new_cell_group.get_groups('ff')
    if not ff_groups:
        ff_group = Group('ff')
        new_cell_group.groups.append(ff_group)
    else:
        ff_group = ff_groups[0]

    # Store content of 'ff' group.
    ff_group.args = [str(cell_type.internal_state), str(cell_type.internal_state) + "_INV"]
    ff_group.set_boolean_function('clocked_on', cell_type.clocked_on)
    ff_group.set_boolean_function('next_state', cell_type.next_state)
    if cell_type.async_preset:
        ff_group.set_boolean_function('preset', cell_type.async_preset)
    if cell_type.async_clear:
        ff_group.set_boolean_function('clear', cell_type.async_clear)

    # Find clock pin.
    clock_signals = list(cell_type.clocked_on.atoms(sympy.Symbol))
    if len(clock_signals) != 1:
        logger.error(f"Expect exactly one clock signal. Got {clock_signals}")
    clock_signal = clock_signals[0]
    # Find clock polarity:
    clock_edge_polarity = cell_type.clocked_on.subs({clock_signal: True})
    clock_pin = str(clock_signal.name)

    assert isinstance(clock_pin, str)

    logger.info(f"Clock signal: {clock_pin}")
    logger.info(f"Clock polarity: {'rising' if clock_edge_polarity else 'falling'}")

    # Find preset/clear signals.
    # Make sure preset/clear are disabled.
    preset_condition = cell_type.async_preset
    clear_condition = cell_type.async_clear

    assert preset_condition is not None, "preset condition must not be `None`" 
    assert clear_condition is not None, "clear condition must not be `None`"

    # Find a variable assignment such that neither preset nor clear is active.
    no_preset_no_clear = list(satisfiable(~preset_condition & ~clear_condition, all_models=True))
    for model in no_preset_no_clear:
        logger.info(f"FF in normal operation mode when: {model}")
    preset_clear_input = no_preset_no_clear[0]
    if len(no_preset_no_clear) > 1:
        logger.warning(f"Multiple possibilities found for disabling preset and clear. "
                       f"Take the first one ({preset_clear_input}).")

    # Find all data pins that are relevant for the internal state of the flip-flop.
    data_in_pins = sorted(cell_type.next_state.atoms(sympy.Symbol))
    logger.debug(f"Input pins relevant for internal state: {data_in_pins}")

    assert isinstance(cell_type.internal_state,
                      sympy.Symbol), "Internal flip-flop-state variable is not defined."

    # Find all output pins that depend on the internal state.
    data_out_pins: List[sympy.Symbol] = [name for name, output in cell_type.outputs.items()
                                         if cell_type.internal_state in output.function.atoms()
                                         ]
    logger.debug(f"Output pins that depend on the internal state: {data_out_pins}")

    
    # Remember a clock pulse width which for sure samples a stable data signal.
    clock_pulse_width_guess = 0

    # TODO use this function to simplify the below code.
    ## Find data input pins which can control the output pin.
    #control_and_observe_pins = find_flipflop_state_control_and_observe_pins(cell_type, preset_clear_input)
    #assert len(control_and_observe_pins) > 0, "No input pin found which can control the flip-flop state."

    static_inputs_and_observer_outputs = seq_recognition.find_flipflop_state_control_and_observe_pins(
                                    cell_type, preset_clear_input
                                )
    
    # Characterize setup/hold for each data pin.
    for i, data_in_pin in enumerate(data_in_pins):
        logger.info(f"Measure constraints of pin {data_in_pin} ({i}/{len(data_in_pins)}).")
        
        for input_pin, static_inputs, observer_outputs, when_other_inputs in static_inputs_and_observer_outputs:
            if input_pin != str(data_in_pin):
                continue

            if not observer_outputs:
                # When the internal state is not observable we cannot measure constraints.
                # Skip this combination.
                logger.warning(
                    f"Internal memory state {cell_type.internal_state} is not observable from any output "
                    f"when {static_inputs}. Skipping input combination.")
                continue
        
            # Take just one of the observer output pins.
            data_out_pin = observer_outputs[0]
            logger.debug(f"Output pin: {data_out_pin}")

            # == Start characterization ==

            # Convert from sympy.Symbol to string.
            data_in_pin = str(data_in_pin)
            data_out_pin = str(data_out_pin)

            # Convert boolean values to input voltages.
            static_input_voltages = {
                pin: cell_conf.global_conf.supply_voltage if value==True else 0 
                for pin, value in static_inputs.items()
            }

            def find_min_clock_pulse_width(clock_pulse_polarity: bool, rising_data_edge: bool):
                min_clock_pulse_width, delay = seq.find_minimum_clock_pulse_width(
                    simulation=simulation,
                    cell_config=cell_conf,
                    ff_clock_edge_polarity=clock_edge_polarity,
                    clock_input=clock_pin,
                    data_in=data_in_pin,
                    data_out=data_out_pin,
                    setup_time=2e-9,  # Choose a reasonably large setup time.
                    clock_pulse_polarity=clock_pulse_polarity,
                    rising_data_edge_in=rising_data_edge,
                    clock_rise_time=10e-12,  # TODO: Something fails when 0.
                    clock_fall_time=10e-12,  # TODO: How to choose this?
                    output_load_capacitances={data_out_pin: 0},
                    clock_pulse_width_guess=100e-12,
                    max_delay_estimation=1e-7,
                    static_input_voltages=static_input_voltages,
                )
                logger.debug(f'min_clock_pulse_width = {min_clock_pulse_width}, delay = {delay}')
                return min_clock_pulse_width, delay

            # Find the minimal clock pulse for negative and positive pulses.
            # For each pulse type inspect rising and falling data edges.
            logger.info(f"Find minimal clock pulse width ({clock_pin}).")
            min_pulse_width_low, _delay = max(find_min_clock_pulse_width(False, False),
                                              find_min_clock_pulse_width(False, True))
            logger.info(f"min_pulse_width_low = {min_pulse_width_low} s")
            min_pulse_width_high, _delay = max(find_min_clock_pulse_width(True, False),
                                               find_min_clock_pulse_width(True, True))
            logger.info(f"min_pulse_width_high = {min_pulse_width_high} s")
            
            # Find the shortest clock pulse width which samles a stable data signal for all cases.
            clock_pulse_width_guess = max(clock_pulse_width_guess, min_pulse_width_low, min_pulse_width_high)

            # Write information on clock pin to liberty.
            # TODO: minimum pulse width is potentially computed for many different input combinations. Take min/max of them! (Now just the last one will be stored)
            clock_pin_group = new_cell_group.get_group('pin', clock_pin)
            clock_pin_group['clock'] = 'true'
            clock_pin_group['min_pulse_width_high'] = min_pulse_width_high / conf.time_unit
            clock_pin_group['min_pulse_width_low'] = min_pulse_width_low / conf.time_unit

            # Find setup and hold times.
            result = seq.characterize_flip_flop_setup_hold(
                simulation=simulation,
                cell_conf=cell_conf,
                data_in_pin=data_in_pin,
                data_out_pin=data_out_pin,
                clock_pin=clock_pin,
                clock_edge_polarity=clock_edge_polarity,

                constrained_pin_transition=input_transition_times,
                related_pin_transition=related_pin_transition,

                output_load_capacitance=0,  # TODO: Is it accurate to assume zero output load?

                static_input_voltages=static_input_voltages
            )

            # Get the table indices.
            # TODO: get correct index/variable mapping from liberty file.
            index_1 = result['related_pin_transition'] / conf.time_unit
            index_2 = result['constrained_pin_transition'] / conf.time_unit
            # TODO: remember all necessary templates and create template tables.

            input_pin_group = new_cell_group.get_group('pin', data_in_pin)

            clock_edge = 'rising' if clock_edge_polarity else 'falling'

            # Add setup/hold information to the liberty pin group.
            for constraint_type in ['hold', 'setup']:
                template_table = liberty_util.create_constraint_template_table(
                    new_library, constraint_type, len(index_1), len(index_2)
                )
                table_template_name = template_table.args[0]

                rise_constraint = Group('rise_constraint', args=[table_template_name])
                rise_constraint.set_array('index_1', index_1)
                rise_constraint.set_array('index_2', index_2)
                rise_constraint.set_array(
                    'values',
                    result[f'{constraint_type}_rise_constraint'] / conf.time_unit
                )

                fall_constraint = Group('fall_constraint', args=[table_template_name])
                fall_constraint.set_array('index_1', index_1)
                fall_constraint.set_array('index_2', index_2)
                fall_constraint.set_array(
                    'values',
                    result[f'{constraint_type}_fall_constraint'] / conf.time_unit
                )

                timing_group = Group(
                    'timing',
                    attributes=[
                        Attribute('timing_type', f'{constraint_type}_{clock_edge}'),
                        Attribute('related_pin', EscapedString(clock_pin))
                    ],
                    groups=[rise_constraint, fall_constraint]
                )

                if len(when_other_inputs.atoms(sympy.Symbol)) > 0:
                    timing_group.set_boolean_function('when', when_other_inputs)

                input_pin_group.groups.append(timing_group)

    
    # Measure clock-to-output delays.
    for data_out_pin in data_out_pins:
        
        logger.info(f"Measure clock-to-output delay for output '{data_out_pin}'")
        
        output_function = cell_type.outputs[data_out_pin].function
        logger.info(f"Output function: {output_function}")

        
        # Convert output pin to string
        data_out_pin_name = str(data_out_pin)
    
        output_pin_group = new_cell_group.get_group('pin', data_out_pin_name)
        output_pin_group.set_boolean_function('function', output_function)
        related_pin = clock_pin
        
        assert len(static_inputs_and_observer_outputs) > 0, "No input pin found which can control the flip-flop state."
            
        # Select an input pin and a combination of other input values such that the input pin controls output of the flip-flop.
        data_in_pin, static_input_values, observer_outputs, _when_other_inputs = static_inputs_and_observer_outputs[0]

        assert data_out_pin in observer_outputs, f"Flip-flop state is not observable from {data_out_pin} when {static_input_values}."
        
        # Determine if this output is the inverse of the internal state given the current static assignments of the other inputs.
        assignment = static_input_values.copy()
        assignment.update({cell_type.internal_state: False, cell_type.internal_state_n: True})
        is_output_inverted = output_function.subs({var: val for var, val in assignment.items() if var is not None})

        # Determine if the non-inverted internal state is really not inverted (it might be inverted).
        input0 = static_input_values.copy()
        input1 = static_input_values.copy()
        input0[data_in_pin] = False
        input1[data_in_pin] = True
        
        next_state_input0 = cell_type.next_state.subs(input0)
        next_state_input1 = cell_type.next_state.subs(input1)

        # Sanity check.
        assert next_state_input0 != next_state_input1, f"state is not controllable by input {data_in_pin} with {static_input_values}" # Should not happen because the input values are chosen such that the state is controllable.
            
        # If data is 0 and state becomes 1, then the state is inverted.
        is_state_inverted = next_state_input0 
        
        # Determine if the current output is the inverse of the data input.
        inverted_output = is_state_inverted ^ is_output_inverted
        assert inverted_output in [sympy.false, sympy.true], f"Can't determine inversion of output from the output function: {output_function}"
        # Convert to bool.
        inverted_output = bool(inverted_output)

        # Translate logic values to voltages.
        static_input_voltages = {
            pin: value * cell_conf.global_conf.supply_voltage 
            for pin, value in static_input_values.items()
        }

        logger.info(f"Use input pin {data_in_pin} with other pin values {static_input_voltages}")
     
        def delay_f(clock_transition_time, output_capacitance, data_edge_polarity: bool) -> float:
            output_load_capacitances = {data_out_pin_name: output_capacitance} 
            
            # Choose setup and hold time to be used for the delay measurement.
            settings = {
                # TODO: use measured setup/hold times
                CalcMode.BEST: (1e-9, 1e-9), # Use large setup and hold times. This leads to shorter propagation time.
                CalcMode.WORST: (1e-9, 1e-9), # Use short setup and hold times. This leads to longer propagation times.
                CalcMode.TYPICAL: (1e-9, 1e-9),
            }
            
            (setup_time, hold_time) = settings[conf.timing_corner]
            
            clock_cycle_hint = clock_pulse_width_guess * 2 # Choose a big-enough clock pulse such that data is sampled for sure.
            data_transition_time = 1e-13 # TODO
            
            delay, slew_time = seq.measure_clock_to_output_delay(
                simulation,
                cell_conf,
                data_in_pin,
                data_out_pin_name,
                clock_pin,
                clock_edge_polarity,
                data_edge_polarity,
                output_load_capacitances,
                clock_transition_time,
                data_transition_time,
                setup_time,
                hold_time,
                clock_cycle_hint,
                static_input_voltages,
                inverted_output=inverted_output,
            )
            return delay, slew_time
      
        def delay_rise_fall(clock_transition_time, output_capacitance) -> Tuple[float, float]:
            rise_delay, rise_transition = delay_f(clock_transition_time, output_capacitance, True)
            fall_delay, fall_transition = delay_f(clock_transition_time, output_capacitance, False)
            return rise_delay, rise_transition, fall_delay, fall_transition        

        f_vec = np.vectorize(delay_rise_fall, cache=True)
        xx, yy = np.meshgrid(input_transition_times, output_capacitances)
        # Evaluate timing on the grid.
        rise_delay, rise_transition, fall_delay, fall_transition = f_vec(xx, yy)

        result = {
            "cell_rise": rise_delay,
            "cell_fall": fall_delay,
            "rise_transition": rise_transition,
            "fall_transition": fall_transition,
            "total_output_net_capacitance": output_capacitances,
            "input_net_transition": input_transition_times
        }

        # Get the table indices.
        # TODO: get correct index/variable mapping from liberty file.
        index_1 = result['total_output_net_capacitance'] / conf.capacitance_unit
        index_2 = result['input_net_transition'] / conf.time_unit

        # Create template tables.
        delay_template_table = liberty_util.create_delay_template_table(new_library, len(index_1), len(index_2))
        delay_table_template_name = delay_template_table.args[0]

        # Create liberty timing tables.
        timing_tables = []
        for table_name in ['cell_rise', 'cell_fall', 'rise_transition', 'fall_transition']:
            table = Group(
                table_name,
                args=[delay_table_template_name],
            )

            table.set_array('index_1', index_1)
            table.set_array('index_2', index_2)
            table.set_array('values', result[table_name] / conf.time_unit)

            timing_tables.append(table)

        # Create the liberty timing group.
        timing_attributes = [
            Attribute('related_pin', EscapedString(related_pin)),
            Attribute('timing_sense', 'non_unate')
        ]

        timing_group = Group(
            'timing',
            attributes=timing_attributes,
            groups=timing_tables
        )

        # Attach timing group to output pin group.
        output_pin_group.groups.append(timing_group)
 



# Copyright (c) 2019-2020 Thomas Kramer.
# SPDX-FileCopyrightText: 2022 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

from itertools import product
from enum import Enum
import inspect
import sympy
from typing import *


class Unate(Enum):
    NON_UNATE = 0
    NEGATIVE_UNATE = -1
    POSITIVE_UNATE = 1


def is_unate_in_xi(bool_function, param_name: str) -> Unate:
    """ Test if the boolean function is unate in input `i`.
    A function `f` is positive unate iff: f(..., x_i = 1, ...) >= f(..., x_i = 0, ...)
    A function `f` is negative unate iff: f(..., x_i = 1, ...) <= f(..., x_i = 0, ...)
    `f` is non-unate or binate if it is neither positive unate nor negative unate.
    :param bool_function:
    :param param_name: Name of input variable.
    :return: Unate.NON_UNATE, Unate.NEGATIVE_UNATE or Unate.POSITIVE_UNATE
    """

    # TODO: use sympy to do this more elegantly (satisfiability solver)

    # Get number of inputs to function.

    params = list(inspect.signature(bool_function).parameters)
    assert param_name in params, "Function does not have parameter with name '{}'.".format(param_name)
    variable_params = [p for p in params if p != param_name]
    num_inputs = len(variable_params)

    is_positive_unate = True
    is_negative_unate = True

    fixed_inputs = list(product(*([[0, 1]] * num_inputs)))

    for inp in fixed_inputs:
        inp = list(inp)

        param_values = {n: v for n, v in zip(variable_params, inp)}
        param_values[param_name] = 0

        out0 = bool_function(**param_values)
        param_values[param_name] = 1
        out1 = bool_function(**param_values)

        _is_negative_unate = out1 <= out0
        _is_positive_unate = out1 >= out0

        is_positive_unate = is_positive_unate and _is_positive_unate
        is_negative_unate = is_negative_unate and _is_negative_unate

        is_non_unate = not is_positive_unate and not is_negative_unate
        if is_non_unate:
            break

    is_non_unate = is_positive_unate == is_negative_unate
    if is_non_unate:
        return Unate.NON_UNATE
    elif is_negative_unate:
        return Unate.NEGATIVE_UNATE
    elif is_positive_unate:
        return Unate.POSITIVE_UNATE
    else:
        assert False


def test_is_unate():
    AND = lambda a, b: a and b
    NAND = lambda a, b: not (a and b)
    XOR = lambda a, b: a ^ b

    assert is_unate_in_xi(AND, 'a') == Unate.POSITIVE_UNATE
    assert is_unate_in_xi(AND, 'b') == Unate.POSITIVE_UNATE
    assert is_unate_in_xi(NAND, 'a') == Unate.NEGATIVE_UNATE
    assert is_unate_in_xi(NAND, 'b') == Unate.NEGATIVE_UNATE

    assert is_unate_in_xi(XOR, 'a') == Unate.NON_UNATE
    
def find_unateness_conditions(f: sympy.Expr, input_variable: sympy.Symbol) -> Tuple[Dict, Dict]:
    """
    Find two sets of variable assignments `p` and `n` such 
    that `f` is positive unate under the assigmnent `p` and negative unate
    under the assigment `n`.

    Returns: (p, n)
    """

    def greater_or_equal(a, b):
        """ Boolean greater-or-equal """
        return (a & ~b) | (a == b)

    def less_or_equal(a, b):
        """ Boolean less-or-equal """
        return greater_or_equal(b, a)
    
    # A function `f` is positive unate iff: f(..., x_i = 1, ...) >= f(..., x_i = 0, ...)
    # A function `f` is negative unate iff: f(..., x_i = 1, ...) <= f(..., x_i = 0, ...)

    f0 = f.subs({input_variable: False})         
    f1 = f.subs({input_variable: True})       
    
    p = list(sympy.satisfiable(greater_or_equal(f1, f0), all_models=True))
    n = list(sympy.satisfiable(less_or_equal(f1, f0), all_models=True))

    return p, n

def test_find_unateness_conditions():

    a, b, c = sympy.symbols('a b c')
    f = a ^ b ^ c

    p, n = find_unateness_conditions(f, a)
    assert len(p) > 0
    assert len(n) > 0

    for cond in p:
        f_ = f.subs(cond)
        assert bool(f_.subs({a: True})) >= bool(f_.subs({a: False})), "not positive unate"
        
    for cond in n:
        f_ = f.subs(cond)
        assert bool(f_.subs({a: True})) <= bool(f_.subs({a: False})), "not negative unate"




# Copyright (c) 2019-2020 Thomas Kramer.
# SPDX-FileCopyrightText: 2022 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
Measurement of the input capacitance by driving the input pin with a constant current.
"""

import os

from itertools import product
from collections import Counter

from .util import *
from .piece_wise_linear import *
from .ngspice_subprocess import CellSimulation, NgSpiceInteractive
import logging
from typing import Dict

from scipy import interpolate

logger = logging.getLogger(__name__)


def characterize_input_capacitances(
        simulator: NgSpiceInteractive,
        input_pins: List[str],
        active_pin: str,
        output_pins: List[str],
        cell_conf: CellConfig
) -> Dict[str, float]:
    """
    Estimate the input capacitance of the `active_pin`.
    The estimation is done by simulating a constant current flowing into an input and measuring the
    time it takes for the input to go from high to low or low to high. This time multiplied by the current
    yields the transported charge which together with the voltage difference tells the capacitance.
    The measurement is done for all combinations of static inputs (all other inputs that are not measured).

    :param input_pins: List of all input pin names.
    :param active_pin: Name of the pin to be measured.
    :param output_pins: List of cell output pins.
    :param config: Parameters for the characterization.

    :returns: Dictionary containing input capacitances for rise, fall and average case.
    """

    assert isinstance(cell_conf, CellConfig)
    assert isinstance(cell_conf.global_conf, CharacterizationConfig)

    # Check for duplicated pins
    def duplicates(items: List) -> List:
        counted = Counter(items)
        return [i for i, c in counted.items() if c > 1]

    assert not duplicates(input_pins), f"Duplicated input pins: {duplicates(input_pins)}"
    assert not duplicates(output_pins), f"Duplicated output pins: {duplicates(output_pins)}"
    
    cfg = cell_conf.global_conf

    inputs_inverted = cell_conf.complementary_pins.values()
    assert active_pin not in inputs_inverted, f"Active pin '{active_pin}' must not be an inverted pin of a differential pair."
    input_pins_non_inverted = [p for p in input_pins if p not in inputs_inverted]
    active_pin_inverted = cell_conf.complementary_pins.get(active_pin)

    logger.debug("characterize_input_capacitances()")
    # Find ports of the SPICE netlist.
    ports = cell_conf.spice_ports
    logger.debug("Subcircuit ports: {}".format(", ".join(ports)))

    logger.debug("Ground net: {}".format(cell_conf.ground_net))
    logger.debug("Supply net: {}".format(cell_conf.supply_net))

    vdd = cfg.supply_voltage
    logger.debug("Vdd: {} V".format(vdd))

    # Create a list of include files.
    setup_statements = cfg.setup_statements + [f".include {cell_conf.spice_netlist_file}"]

    # Load include files.
    for setup in setup_statements:
        logger.debug(f"Setup statement: {setup}")
    setup_statements_string = "\n".join(setup_statements)

    # Add output load capacitance. Right now this is 1fF.
    output_load_statements = "\n".join((f"Cload_{p} {p} {cell_conf.ground_net} 1f" for p in output_pins))
        
    # Choose a maximum time to run the simulation.
    time_max = cfg.time_step * 1e6

    # Find function to summarize different timing arcs.
    reduction_function = {
        CalcMode.WORST: max,
        CalcMode.BEST: min,
        CalcMode.TYPICAL: np.mean
    }[cfg.timing_corner]
    logger.debug("Reduction function for summarizing multiple timing arcs: {}".format(reduction_function.__name__))

    logger.debug("Measuring input capacitance.")

    # If the ground net is not called GND, then it must be connected to GND or 0.
    # This seems to be a must for SPICE simulators.
    # See: https://codeberg.org/librecell/lctime/issues/12#issuecomment-2173526
    if cell_conf.ground_net not in ["GND", "gnd", "0"]:
        logger.debug(f"Connect ground net `{cell_conf.ground_net}` to simulators ground `0`.")
        ground_connection = f"* Connect ground net to GND\n" \
        f"vgnd {cell_conf.ground_net} 0 0"
    else:
        # Using standard name for ground net.
        ground_connection = ""


    # Generate all possible input combinations for the static input pins.
    static_input_nets = [i for i in input_pins_non_inverted if i != active_pin]
    num_inputs = len(static_input_nets)

    static_inputs = list(product(*([[0, 1]] * num_inputs)))
    logger.debug(f"Number of static input combinations: {len(static_inputs)}")

    input_current = cfg.input_current_for_capacitance_measurement
    logger.debug("Input current: {}".format(input_current))

    simulator.reset()

    input_voltages = dict()
    input_currents = dict()

    for pin in input_pins:
        if pin != active_pin:
            input_voltages[pin] = 0.0
    input_currents[active_pin] = 0.0
    if active_pin_inverted:
        input_current[active_pin_inverted] = 0.0
            
    input_voltages[cell_conf.supply_net] = cell_conf.global_conf.supply_voltage
    if cell_conf.pwell_pin:
        input_voltages[cell_conf.pwell_pin] = 0.0
    if cell_conf.nwell_pin:
        input_voltages[cell_conf.nwell_pin] = cell_conf.global_conf.supply_voltage
    # Create a list of include files.
    setup_statements = cell_conf.global_conf.setup_statements + [f".include {cell_conf.spice_netlist_file}"]

    initial_node_voltages = {active_pin: 0.0}
    if active_pin_inverted:
        initial_node_voltages[active_pin_inverted] = 0.0
    
    # Create simulation harness.
    simulation = CellSimulation(
        simulator,
        cell_name=cell_conf.cell_name,
        cell_ports=cell_conf.spice_ports,
        input_voltages=input_voltages, # First set all to 0V.
        input_currents=input_currents,
        output_load_capacitances={p: 1e-15 for p in output_pins}, # TODO
        initial_node_voltages=initial_node_voltages,
        setup_statements=setup_statements,
        ground_net=cell_conf.ground_net,
        simulation_title=f"measure input capacitance of {cell_conf.cell_name} {active_pin}",
        # Let ngspice find the initial voltages for output pins.
        set_initial_voltages_for_outputs=False
    )


    # Loop through all combinations of inputs.
    capacitances_rising = []
    capacitances_falling = []
    for static_input in static_inputs:
        for input_rising in [True, False]:
            simulation.reset()

            # Get voltages at static inputs.
            input_voltages = {net: cfg.supply_voltage * value for net, value in zip(static_input_nets, static_input)}
            # Set supply voltage.
            input_voltages[cell_conf.supply_net] = cfg.supply_voltage

            # Add input voltages for inverted inputs of differential pairs.
            for p in static_input_nets:
                inv = cell_conf.complementary_pins.get(p)
                if inv is not None:
                    assert inv not in input_voltages
                    # Add the inverted input voltage.
                    input_voltages[inv] = cfg.supply_voltage - input_voltages[p]

            # Add nwell voltage
            if cell_conf.nwell_pin is not None:
                input_voltages[cell_conf.nwell_pin] = cfg.supply_voltage
                
            # Add pwell voltage
            if cell_conf.pwell_pin is not None:
                input_voltages[cell_conf.pwell_pin] = 0.0
                
            logger.debug("Static input voltages: {}".format(input_voltages))

            # Simulation script file path.
            file_name = f"lctime_input_capacitance_" \
                        f"{''.join((f'{net}={v}' for net, v in input_voltages.items()))}_" \
                        f"{'rising' if input_rising else 'falling'}"

            # File for debug plot of the waveforms.
            sim_plot_file = os.path.join(cfg.workingdir, f"{file_name}_plot.svg")

            # Switch polarity of current for falling edges.
            _input_current = input_current if input_rising else -input_current

            # Get initial voltage of active pin.
            initial_voltage = 0 if input_rising else vdd
            initial_voltage_inv = vdd - initial_voltage

            # Get the breakpoint condition.
            if input_rising:
                breakpoint_statement = f"stop when v({active_pin}) > {vdd * 0.9}"
            else:
                breakpoint_statement = f"stop when v({active_pin}) < {vdd * 0.1}"
                
            # Initial node voltages.
            # Add static input voltages
            initial_conditions = input_voltages.copy()
            initial_conditions[active_pin] = initial_voltage
            initial_conditions[cell_conf.supply_net] = cfg.supply_voltage

            # Add initial voltage of inverted input pin (if any).
            if active_pin_inverted:
                initial_conditions[active_pin_inverted] = initial_voltage_inv

            for net, voltage in initial_conditions.items():
                simulation.set_initial_voltage(net, voltage)
            simulation.simulator.reset()

            # Set input voltages.
            for net, value in input_voltages.items():
                src = f"V{net}"
                simulation.set_source(src, value)

            # Set input current.
            ps = 1e-15
            ns = 1e-9
            input_current_pwl = PieceWiseLinear(
                [1*ns, 1*ns+1*ps],
                [0.0, _input_current],
                unit_y=""
            )

            simulation.set_source(f"I{active_pin}", input_current_pwl)
            if active_pin_inverted:
                simulation.set_source(f"I{active_pin_inverted}", -input_current_pwl)

            simulation.set_breakpoint(breakpoint_statement)

            # Set output loads.
            for pin in output_pins:
                simulation.set_capacitance(f"Cload_{pin}", 1e-15)

            time, voltages, currents = simulation.tran(
                t_step=cfg.time_step, t_stop=time_max,
                output_voltages=[active_pin],
                output_currents=[],
            )

            input_voltage = voltages[active_pin]

            if cfg.debug_plots:
                logger.info("Create plot of waveforms: {}".format(sim_plot_file))
                import matplotlib
                matplotlib.use('Agg')
                import matplotlib.pyplot as plt
                plt.close()
                plt.title(f"Measure input capacitance of pin {active_pin}.")
                plt.plot(time, input_voltage, label=f"V({active_pin}")
                plt.legend()
                plt.savefig(sim_plot_file)
                plt.close()

            # Calculate average derivative of voltage by finding the slope of the line
            # through the crossing point of the voltage with the two thresholds.
            #
            # TODO: How to chose the thresholds?
            if input_rising:
                thresh1 = vdd * cfg.trip_points.slew_lower_threshold_rise
                thresh2 = vdd * cfg.trip_points.slew_upper_threshold_rise
                assert thresh1 < thresh2
            else:
                thresh1 = vdd * cfg.trip_points.slew_upper_threshold_fall
                thresh2 = vdd * cfg.trip_points.slew_lower_threshold_fall
                assert thresh1 > thresh2

            # Find transition times for both thresholds.
            transition_time1 = transition_time(input_voltage, time, threshold=thresh1, n=-1)
            transition_time2 = transition_time(input_voltage, time, threshold=thresh2, n=-1)
            assert transition_time2 > transition_time1

            # Compute deltas of time and voltage between the crossing of the two thresholds.
            f_input_voltage = interpolate.interp1d(x=time, y=input_voltage)
            dt = transition_time2 - transition_time1
            dv = f_input_voltage(transition_time2) - f_input_voltage(transition_time1)

            # Compute capacitance.
            capacitance = float(_input_current) / (float(dv) / float(dt))

            logger.debug("dV: {}".format(dv))
            logger.debug("dt: {}".format(dt))
            logger.debug("I: {}".format(input_current))
            logger.debug("Input capacitance {}: {} F".format(active_pin, capacitance))

            if input_rising:
                capacitances_rising.append(capacitance)
            else:
                capacitances_falling.append(capacitance)

    logger.debug("Characterizing input capacitances: Done")

    # Find max, min or average depending on 'reduction_function'.
    logger.debug(
        "Convert capacitances of all timing arcs into the default capacitance ({})".format(reduction_function.__name__))
    final_capacitance_falling = reduction_function(capacitances_falling)
    final_capacitance_rising = reduction_function(capacitances_rising)
    final_capacitance = reduction_function([final_capacitance_falling, final_capacitance_rising])

    return {
        'rise_capacitance': final_capacitance_falling,
        'fall_capacitance': final_capacitance_rising,
        'capacitance': final_capacitance
    }
